@@ -1,5 +1,5 @@
 /*
-    $Id: addressobj.c 2122 2019-12-21 06:27:50Z soci $
+    $Id: addressobj.c 2965 2023-01-08 19:50:05Z soci $
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -22,23 +22,24 @@
 #include "eval.h"
 #include "variables.h"
 #include "arguments.h"
+#include "instruction.h"
 
 #include "boolobj.h"
 #include "strobj.h"
 #include "intobj.h"
-#include "operobj.h"
 #include "typeobj.h"
 #include "noneobj.h"
 #include "errorobj.h"
 #include "floatobj.h"
 #include "bitsobj.h"
 #include "bytesobj.h"
+#include "registerobj.h"
 
 static Type obj;
 
 Type *const ADDRESS_OBJ = &obj;
 
-static MUST_CHECK Obj *create(Obj *v1, linepos_t epoint) {
+static MUST_CHECK Obj *address_from_obj(Obj *v1, linepos_t epoint) {
     switch (v1->obj->type) {
     case T_BOOL:
     case T_INT:
@@ -46,22 +47,26 @@ static MUST_CHECK Obj *create(Obj *v1, linepos_t epoint) {
     case T_FLOAT:
     case T_BYTES:
     case T_STR:
-        return (Obj *)new_address(val_reference(v1), A_NONE);
+        return new_address(val_reference(v1), A_NONE);
     case T_NONE:
     case T_ERROR:
     case T_ADDRESS: return val_reference(v1);
     default: break;
     }
-    return (Obj *)new_error_conv(v1, ADDRESS_OBJ, epoint);
+    return new_error_conv(v1, ADDRESS_OBJ, epoint);
+}
+
+static MUST_CHECK Obj *convert(oper_t op) {
+    return address_from_obj(op->v2, op->epoint2);
 }
 
 static FAST_CALL void destroy(Obj *o1) {
-    Address *v1 = (Address *)o1;
+    Address *v1 = Address(o1);
     val_destroy(v1->val);
 }
 
 static FAST_CALL void garbage(Obj *o1, int i) {
-    Address *v1 = (Address *)o1;
+    Address *v1 = Address(o1);
     Obj *v;
     switch (i) {
     case -1:
@@ -80,24 +85,24 @@ static FAST_CALL void garbage(Obj *o1, int i) {
 }
 
 static FAST_CALL bool same(const Obj *o1, const Obj *o2) {
-    const Address *v1 = (const Address *)o1, *v2 = (const Address *)o2;
-    return o2->obj == ADDRESS_OBJ && v1->type == v2->type && v1->val->obj->same(v1->val, v2->val);
+    const Address *v1 = Address(o1), *v2 = Address(o2);
+    return o1->obj == o2->obj && v1->type == v2->type && v1->val->obj->same(v1->val, v2->val);
 }
 
 static MUST_CHECK Obj *truth(Obj *o1, Truth_types type, linepos_t epoint) {
-    Address *v1 = (Address *)o1;
+    Address *v1 = Address(o1);
     Obj *v;
-    if (v1->type != A_NONE && v1->val != &none_value->v && v1->val->obj != ERROR_OBJ) {
+    if (v1->type != A_NONE && v1->val != none_value && v1->val->obj != ERROR_OBJ) {
         return DEFAULT_OBJ->truth(o1, type, epoint);
     }
     v = v1->val;
     return v->obj->truth(v, type, epoint);
 }
 
-static MUST_CHECK Error *hash(Obj *o1, int *hs, linepos_t epoint) {
-    Address *v1 = (Address *)o1;
+static MUST_CHECK Obj *hash(Obj *o1, int *hs, linepos_t epoint) {
+    Address *v1 = Address(o1);
     Obj *v = v1->val;
-    Error *err = v->obj->hash(v, hs, epoint);
+    Obj *err = v->obj->hash(v, hs, epoint);
     if (err == NULL) {
         *hs = ((unsigned int)*hs + v1->type) & ((~0U) >> 1);
     }
@@ -105,7 +110,7 @@ static MUST_CHECK Error *hash(Obj *o1, int *hs, linepos_t epoint) {
 }
 
 static MUST_CHECK Obj *repr(Obj *o1, linepos_t epoint, size_t maxsize) {
-    Address *v1 = (Address *)o1;
+    Address *v1 = Address(o1);
     uint8_t *s;
     size_t len, chars;
     char buffer[100], buffer2[100], *b2;
@@ -157,9 +162,8 @@ static MUST_CHECK Obj *repr(Obj *o1, linepos_t epoint, size_t maxsize) {
     if (chars > maxsize) return NULL;
     tmp = v1->val->obj->repr(v1->val, epoint, maxsize - chars);
     if (tmp == NULL || tmp->obj != STR_OBJ) return tmp;
-    str = (Str *)tmp;
-    len = chars + str->len;
-    if (len < chars) goto error; /* overflow */
+    str = Str(tmp);
+    if (add_overflow(str->len, chars, &len)) goto error;
     chars += str->chars;
     if (chars > maxsize) {
     error:
@@ -181,11 +185,11 @@ static MUST_CHECK Obj *repr(Obj *o1, linepos_t epoint, size_t maxsize) {
     }
     if (ind2 != 0) memcpy(s, buffer, ind2);
     val_destroy(tmp);
-    return &v->v;
+    return Obj(v);
 }
 
 static MUST_CHECK Obj *str(Obj *o1, linepos_t epoint, size_t maxsize) {
-    Address *v1 = (Address *)o1;
+    Address *v1 = Address(o1);
     if (v1->type == A_NONE) {
         o1 = v1->val;
         return o1->obj->repr(o1, epoint, maxsize);
@@ -237,8 +241,22 @@ static inline bool check_addr2(atype_t type) {
     return false;
 }
 
+Address_types register_to_indexing(unsigned int c) {
+    switch (c) {
+    case 's': return A_SR;
+    case 'r': return A_RR;
+    case 'z': return A_ZR;
+    case 'y': return A_YR;
+    case 'x': return A_XR;
+    case 'd': return A_DR;
+    case 'b': return A_BR;
+    case 'k': return A_KR;
+    default: return A_NONE;
+    }
+}
+
 static FAST_CALL uint32_t address(const Obj *o1) {
-    const Address *v1 = (Address *)o1;
+    const Address *v1 = Address(o1);
     Obj *v = v1->val;
     uint32_t am = v->obj->address(v);
     atype_t type = v1->type;
@@ -250,9 +268,9 @@ static FAST_CALL uint32_t address(const Obj *o1) {
 }
 
 static MUST_CHECK Error *ival(Obj *o1, ival_t *iv, unsigned int bits, linepos_t epoint) {
-    Address *v1 = (Address *)o1;
+    Address *v1 = Address(o1);
     Obj *v;
-    if (v1->type != A_NONE && v1->val != &none_value->v && v1->val->obj != ERROR_OBJ) {
+    if (v1->type != A_NONE && v1->val != none_value && v1->val->obj != ERROR_OBJ) {
         return DEFAULT_OBJ->ival(o1, iv, bits, epoint);
     }
     v = v1->val;
@@ -260,9 +278,9 @@ static MUST_CHECK Error *ival(Obj *o1, ival_t *iv, unsigned int bits, linepos_t
 }
 
 static MUST_CHECK Error *uval(Obj *o1, uval_t *uv, unsigned int bits, linepos_t epoint) {
-    Address *v1 = (Address *)o1;
+    Address *v1 = Address(o1);
     Obj *v;
-    if (v1->type != A_NONE && v1->val != &none_value->v && v1->val->obj != ERROR_OBJ) {
+    if (v1->type != A_NONE && v1->val != none_value && v1->val->obj != ERROR_OBJ) {
         return DEFAULT_OBJ->uval(o1, uv, bits, epoint);
     }
     v = v1->val;
@@ -270,9 +288,9 @@ static MUST_CHECK Error *uval(Obj *o1, uval_t *uv, unsigned int bits, linepos_t
 }
 
 static MUST_CHECK Error *uval2(Obj *o1, uval_t *uv, unsigned int bits, linepos_t epoint) {
-    Address *v1 = (Address *)o1;
+    Address *v1 = Address(o1);
     Obj *v;
-    if (v1->type != A_NONE && v1->val != &none_value->v && v1->val->obj != ERROR_OBJ) {
+    if (v1->type != A_NONE && v1->val != none_value && v1->val->obj != ERROR_OBJ) {
         return DEFAULT_OBJ->uval2(o1, uv, bits, epoint);
     }
     v = v1->val;
@@ -280,49 +298,49 @@ static MUST_CHECK Error *uval2(Obj *o1, uval_t *uv, unsigned int bits, linepos_t
 }
 
 static MUST_CHECK Error *iaddress(Obj *o1, ival_t *iv, unsigned int bits, linepos_t epoint) {
-    Address *v1 = (Address *)o1;
+    Address *v1 = Address(o1);
     Obj *v = v1->val;
     return v->obj->iaddress(v, iv, bits, epoint);
 }
 
 static MUST_CHECK Error *uaddress(Obj *o1, uval_t *uv, unsigned int bits, linepos_t epoint) {
-    Address *v1 = (Address *)o1;
+    Address *v1 = Address(o1);
     Obj *v = v1->val;
     return v->obj->uaddress(v, uv, bits, epoint);
 }
 
 MUST_CHECK Obj *float_from_address(Address *v1, linepos_t epoint) {
-    if (v1->type != A_NONE && v1->val != &none_value->v && v1->val->obj != ERROR_OBJ) {
-        return (Obj *)new_error_conv(&v1->v, FLOAT_OBJ, epoint);
+    if (v1->type != A_NONE && v1->val != none_value && v1->val->obj != ERROR_OBJ) {
+        return new_error_conv(Obj(v1), FLOAT_OBJ, epoint);
     }
-    return FLOAT_OBJ->create(v1->val, epoint);
+    return float_from_obj(v1->val, epoint);
 }
 
 MUST_CHECK Obj *int_from_address(Address *v1, linepos_t epoint) {
-    if (v1->type != A_NONE && v1->val != &none_value->v && v1->val->obj != ERROR_OBJ) {
-        return (Obj *)new_error_conv(&v1->v, INT_OBJ, epoint);
+    if (v1->type != A_NONE && v1->val != none_value && v1->val->obj != ERROR_OBJ) {
+        return new_error_conv(Obj(v1), INT_OBJ, epoint);
     }
-    return INT_OBJ->create(v1->val, epoint);
+    return int_from_obj(v1->val, epoint);
 }
 
 MUST_CHECK Obj *bits_from_address(Address *v1, linepos_t epoint) {
-    if (v1->type != A_NONE && v1->val != &none_value->v && v1->val->obj != ERROR_OBJ) {
-        return (Obj *)new_error_conv(&v1->v, BITS_OBJ, epoint);
+    if (v1->type != A_NONE && v1->val != none_value && v1->val->obj != ERROR_OBJ) {
+        return new_error_conv(Obj(v1), BITS_OBJ, epoint);
     }
-    return BITS_OBJ->create(v1->val, epoint);
+    return bits_from_obj(v1->val, epoint);
 }
 
 MUST_CHECK Obj *bytes_from_address(Address *v1, linepos_t epoint) {
-    if (v1->type != A_NONE && v1->val != &none_value->v && v1->val->obj != ERROR_OBJ) {
-        return (Obj *)new_error_conv(&v1->v, BYTES_OBJ, epoint);
+    if (v1->type != A_NONE && v1->val != none_value && v1->val->obj != ERROR_OBJ) {
+        return new_error_conv(Obj(v1), BYTES_OBJ, epoint);
     }
-    return BYTES_OBJ->create(v1->val, epoint);
+    return bytes_from_obj(v1->val, epoint);
 }
 
 static MUST_CHECK Obj *sign(Obj *o1, linepos_t epoint) {
-    Address *v1 = (Address *)o1;
+    Address *v1 = Address(o1);
     Obj *v;
-    if (v1->type != A_NONE && v1->val != &none_value->v && v1->val->obj != ERROR_OBJ) {
+    if (v1->type != A_NONE && v1->val != none_value && v1->val->obj != ERROR_OBJ) {
         return DEFAULT_OBJ->sign(o1, epoint);
     }
     v = v1->val;
@@ -330,51 +348,49 @@ static MUST_CHECK Obj *sign(Obj *o1, linepos_t epoint) {
 }
 
 static MUST_CHECK Obj *function(oper_t op) {
-    Address *v1 = (Address *)op->v2;
+    Address *v1 = Address(op->v2);
     Obj *v;
-    if (v1->type != A_NONE && v1->val != &none_value->v && v1->val->obj != ERROR_OBJ) {
+    if (v1->type != A_NONE && v1->val != none_value && v1->val->obj != ERROR_OBJ) {
         return DEFAULT_OBJ->function(op);
     }
     op->v2 = v = v1->val;
-    op->inplace = op->inplace == &v1->v && v->refcount == 1 ? v : NULL;
+    op->inplace = op->inplace == Obj(v1) && v->refcount == 1 ? v : NULL;
     return v->obj->function(op);
 }
 
 static MUST_CHECK Obj *calc1(oper_t op) {
     Obj *result;
-    Address *v1 = (Address *)op->v1;
+    Address *v1 = Address(op->v1);
     atype_t am = v1->type;
-    switch (op->op->op) {
+    switch (op->op) {
     case O_LNOT:
     case O_STRING:
         if (am != A_NONE) break;
-        /* fall through */
+        FALL_THROUGH; /* fall through */
     case O_BANK:
     case O_HIGHER:
     case O_LOWER:
     case O_HWORD:
     case O_WORD:
     case O_BSWORD:
-        goto ok;
     case O_INV:
     case O_NEG:
     case O_POS:
         if (check_addr2(am)) break;
-    ok:
         op->v1 = v1->val;
         op->inplace = NULL;
         result = op->v1->obj->calc1(op);
         if (am == A_NONE) return result;
-        return (Obj *)new_address(result, am);
+        return new_address(result, am);
     default: break;
     }
-    if (v1->val == &none_value->v || v1->val->obj == ERROR_OBJ) return val_reference(v1->val);
+    if (v1->val == none_value || v1->val->obj == ERROR_OBJ) return val_reference(v1->val);
     return obj_oper_error(op);
 }
 
-static MUST_CHECK Obj *slice(oper_t op, size_t indx) {
-    Obj *val = ((Address *)op->v1)->val;
-    if (val == &none_value->v || val->obj == ERROR_OBJ) {
+static MUST_CHECK Obj *slice(oper_t op, argcount_t indx) {
+    Obj *val = Address(op->v1)->val;
+    if (val == none_value || val->obj == ERROR_OBJ) {
         return val_reference(val);
     }
     return DEFAULT_OBJ->slice(op, indx);
@@ -382,23 +398,25 @@ static MUST_CHECK Obj *slice(oper_t op, size_t indx) {
 
 static MUST_CHECK Obj *calc2(oper_t op) {
     Obj *o2 = op->v2, *result;
-    Address *v1 = (Address *)op->v1, *v;
+    Address *v1 = Address(op->v1);
     atype_t am;
-    if (op->op == &o_LAND || op->op == &o_LOR) {
+    if (op->op == O_LAND || op->op == O_LOR || op->op == O_LXOR) {
         bool i;
-        result = truth(&v1->v, TRUTH_BOOL, op->epoint);
+        result = truth(Obj(v1), TRUTH_BOOL, op->epoint);
         if (result->obj != BOOL_OBJ) return result;
-        i = (result == &true_value->v) != (op->op == &o_LOR);
+        i = Bool(result)->value;
         val_destroy(result);
         if (diagnostics.strict_bool) err_msg_bool_oper(op);
-        return val_reference(i ? o2 : &v1->v);
+        if (op->op == O_LXOR) return calc2_lxor(op, i);
+        return val_reference(i != (op->op == O_LOR) ? o2 : Obj(v1));
     }
     switch (o2->obj->type) {
     case T_ADDRESS:
         {
-            Address *v2 = (Address *)o2;
+            Address *v2 = Address(o2);
+            atype_t am2 = v2->type;
             am = v1->type;
-            switch (op->op->op) {
+            switch (op->op) {
             case O_CMP:
             case O_EQ:
             case O_NE:
@@ -408,44 +426,30 @@ static MUST_CHECK Obj *calc2(oper_t op) {
             case O_MAX:
             case O_GT:
             case O_GE:
-                if (am == v2->type) {
+                if (am == am2) {
                     op->v1 = v1->val;
                     op->v2 = v2->val;
                     op->inplace = NULL;
                     return op->v1->obj->calc2(op);
                 }
-                switch (op->op->op) {
-                default: /* can't happen */
-                case O_CMP: return (Obj *)ref_int((am < v2->type) ? minus1_value : int_value[1]);
-                case O_EQ: return (Obj *)ref_bool(false_value);
-                case O_NE: return (Obj *)ref_bool(true_value);
-                case O_LE:
-                case O_MIN:
-                case O_LT: return truth_reference(am < v2->type);
-                case O_GE:
-                case O_MAX:
-                case O_GT: return truth_reference(am > v2->type);
-                }
+                return obj_oper_compare(op, (am < am2) ? -1 : 1);
             case O_ADD:
                 if (check_addr(am)) break;
-                if (check_addr(v2->type)) break;
+                if (check_addr(am2)) break;
                 op->v1 = v1->val;
                 op->v2 = v2->val;
                 op->inplace = NULL;
                 result = op->v1->obj->calc2(op);
                 if (am == A_NONE && v2->type == A_NONE) return result;
-                v = new_address(result, am);
-                am = v2->type;
-                while ((Address_types)(am & 0xf) != A_NONE) { v->type <<= 4; am >>= 4; }
-                v->type |= v2->type;
-                return &v->v;
+                while ((Address_types)(am2 & 0xf) != A_NONE) { am <<= 4; am2 >>= 4; }
+                return new_address(result, am | v2->type);
             case O_SUB:
                 if (check_addr(am)) break;
-                if (check_addr(v2->type)) break;
+                if (check_addr(am2)) break;
                 {
-                    atype_t am1 = A_NONE, am2 = v2->type;
+                    atype_t am1 = A_NONE;
                     for (; (am & MAX_ADDRESS_MASK) != 0; am <<= 4) {
-                        atype_t amc = (am >> 8) & 0xf;
+                        atype_t amc = (am >> 12) & 0xf;
                         atype_t am3, am4;
                         if (amc == A_NONE) continue;
                         am3 = A_NONE; am4 = am2;
@@ -465,23 +469,31 @@ static MUST_CHECK Obj *calc2(oper_t op) {
                     op->inplace = NULL;
                     result = op->v1->obj->calc2(op);
                     if (am1 == A_NONE) return result;
-                    return (Obj *)new_address(result, am1);
+                    return new_address(result, am1);
                 }
+            case O_XOR:
+                if (check_addr(am)) break;
+                if (check_addr(am2)) break;
+                if (am == am2) {
+                    op->v1 = v1->val;
+                    op->v2 = v2->val;
+                    op->inplace = NULL;
+                    return op->v1->obj->calc2(op);
+                }
+                break;
             default:
                 break;
             }
             break;
         }
     case T_BOOL:
-        if (diagnostics.strict_bool) err_msg_bool_oper(op);
-        /* fall through */
     case T_INT:
     case T_BITS:
     case T_FLOAT:
     case T_BYTES:
     case T_STR:
         am = v1->type;
-        switch (op->op->op) {
+        switch (op->op) {
         case O_CMP:
         case O_EQ:
         case O_NE:
@@ -491,7 +503,7 @@ static MUST_CHECK Obj *calc2(oper_t op) {
         case O_MAX:
         case O_GT:
         case O_GE:
-            if (am == A_NONE) {
+            if (am == A_NONE || (am == A_DR && dpage == 0) || (am == A_BR && databank == 0)) {
                 op->v1 = v1->val;
                 op->inplace = NULL;
                 return op->v1->obj->calc2(op);
@@ -511,35 +523,77 @@ static MUST_CHECK Obj *calc2(oper_t op) {
         ok:
             op->v1 = v1->val;
             op->inplace = NULL;
-            return (Obj *)new_address(op->v1->obj->calc2(op), am);
+            return new_address(op->v1->obj->calc2(op), am);
+        }
+        break;
+    case T_REGISTER:
+        if (Register(op->v2)->len == 1) {
+            Address_types am2 = register_to_indexing(Register(op->v2)->data[0]);
+            if (am2 == A_NONE) break;
+            if (op->op == O_ADD) {
+                return new_address(val_reference(v1->val), v1->type << 4 | am2);
+            } 
+            if (op->op == O_SUB) {
+                atype_t am1 = A_NONE;
+                for (am = v1->type; ; am >>= 4) {
+                    atype_t amc = am & 0xf;
+                    switch (amc) {
+                    case A_I:
+                    case A_LI:
+                    case A_IMMEDIATE:
+                    case A_IMMEDIATE_SIGNED: 
+                        if (am2 == A_NONE) {
+                            am1 = (am1 >> 4) | (amc << 12);
+                            continue;
+                        }
+                        break;
+                    case A_KR:
+                    case A_DR:
+                    case A_BR:
+                    case A_XR:
+                    case A_YR:
+                    case A_ZR:
+                    case A_RR:
+                    case A_SR:
+                        if ((Address_types)amc == am2) {
+                            am2 = A_NONE;
+                            continue;
+                        }
+                        am1 = (am1 >> 4) | (amc << 12);
+                        continue;
+                    default: break;
+                    }
+                    break;
+                }
+                if (am2 != A_NONE) break;
+                if (am1 != A_NONE) while ((am1 & 0xf) == A_NONE) am1 >>= 4;
+                return new_address(val_reference(v1->val), am1);
+            }
         }
         break;
     default:
-        if (op->op != &o_MEMBER && op->op != &o_X) {
+        if (op->op != O_MEMBER && op->op != O_X) {
             return o2->obj->rcalc2(op);
         }
-        if (o2 == &none_value->v || o2->obj == ERROR_OBJ) return val_reference(o2);
+        if (o2 == none_value || o2->obj == ERROR_OBJ) return val_reference(o2);
     }
-    if (v1->val == &none_value->v || v1->val->obj == ERROR_OBJ) return val_reference(v1->val);
+    if (v1->val == none_value || v1->val->obj == ERROR_OBJ) return val_reference(v1->val);
     return obj_oper_error(op);
 }
 
 static MUST_CHECK Obj *rcalc2(oper_t op) {
     const Type *t1 = op->v1->obj;
-    Address *v2 = (Address *)op->v2;
+    Address *v2 = Address(op->v2);
     atype_t am;
     switch (t1->type) {
     case T_BOOL:
-        if (diagnostics.strict_bool) err_msg_bool_oper(op);
-        /* fall through */
     case T_INT:
     case T_BITS:
     case T_FLOAT:
     case T_BYTES:
-    case T_STR:
     case T_GAP:
         am = v2->type;
-        switch (op->op->op) {
+        switch (op->op) {
         default:
             if (am == A_NONE) {
                 op->v2 = v2->val;
@@ -558,44 +612,58 @@ static MUST_CHECK Obj *rcalc2(oper_t op) {
         ok:
             op->v2 = v2->val;
             op->inplace = NULL;
-            return (Obj *)new_address(t1->calc2(op), am);
+            return new_address(t1->calc2(op), am);
         }
         break;
     case T_CODE:
-        if (op->op != &o_IN) {
-            return t1->calc2(op);
+        return t1->calc2(op);
+    case T_REGISTER:
+        if (Register(op->v1)->len == 1) {
+            am = register_to_indexing(Register(op->v1)->data[0]);
+            if (am == A_NONE) break;
+            if (op->op == O_ADD) {
+                return new_address(val_reference(v2->val), v2->type << 4 | am);
+            }
+            if (op->op == O_SUB) {
+                if (am == v2->type) {
+                    op->v1 = int_value[0];
+                    op->v2 = v2->val;
+                    op->inplace = NULL;
+                    return INT_OBJ->calc2(op);
+                }
+            }
         }
         break;
     default: break;
     }
-    if (v2->val == &none_value->v || v2->val->obj == ERROR_OBJ) return val_reference(v2->val);
+    if (v2->val == none_value || v2->val->obj == ERROR_OBJ) return val_reference(v2->val);
     return obj_oper_error(op);
 }
 
 void addressobj_init(void) {
-    new_type(&obj, T_ADDRESS, "address", sizeof(Address));
-    obj.create = create;
-    obj.destroy = destroy;
-    obj.garbage = garbage;
-    obj.same = same;
-    obj.truth = truth;
-    obj.hash = hash;
-    obj.repr = repr;
-    obj.str = str;
-    obj.address = address;
-    obj.ival = ival;
-    obj.uval = uval;
-    obj.uval2 = uval2;
-    obj.iaddress = iaddress;
-    obj.uaddress = uaddress;
-    obj.sign = sign;
-    obj.function = function;
-    obj.calc1 = calc1;
-    obj.calc2 = calc2;
-    obj.rcalc2 = rcalc2;
-    obj.slice = slice;
+    Type *type = new_type(&obj, T_ADDRESS, "address", sizeof(Address));
+    type->convert = convert;
+    type->destroy = destroy;
+    type->garbage = garbage;
+    type->same = same;
+    type->truth = truth;
+    type->hash = hash;
+    type->repr = repr;
+    type->str = str;
+    type->address = address;
+    type->ival = ival;
+    type->uval = uval;
+    type->uval2 = uval2;
+    type->iaddress = iaddress;
+    type->uaddress = uaddress;
+    type->sign = sign;
+    type->function = function;
+    type->calc1 = calc1;
+    type->calc2 = calc2;
+    type->rcalc2 = rcalc2;
+    type->slice = slice;
 }
 
 void addressobj_names(void) {
-    new_builtin("address", val_reference(&ADDRESS_OBJ->v));
+    new_builtin("address", val_reference(Obj(ADDRESS_OBJ)));
 }
