@@ -1,5 +1,5 @@
 /*
-    $Id: arguments.h 2125 2019-12-23 19:50:58Z soci $
+    $Id: arguments.h 2954 2023-01-07 10:22:44Z soci $
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -23,7 +23,7 @@
 
 typedef enum Output_types {
     OUTPUT_CBM, OUTPUT_RAW, OUTPUT_NONLINEAR, OUTPUT_FLAT, OUTPUT_XEX,
-    OUTPUT_APPLE, OUTPUT_IHEX, OUTPUT_SREC
+    OUTPUT_APPLE, OUTPUT_IHEX, OUTPUT_SREC, OUTPUT_MHEX
 } Output_types;
 
 typedef enum Symbollist_types {
@@ -34,32 +34,62 @@ typedef enum Caret_types {
     CARET_ALWAYS, CARET_MACRO, CARET_NEVER
 } Caret_types;
 
-struct file_s;
-
 struct output_s {
     const char *name;
     const char *section;
+    const char *mapname;
     Output_types mode;
+    bool append;
     bool longaddr;
+    bool mapappend;
+    bool mapfile;
+};
+
+struct error_output_s {
+    const char *name;
+    Caret_types caret;
+    bool warning;
+    bool no_output;
+    bool append;
 };
 
 struct symbol_output_s {
     const char *name;
-    Symbollist_types mode;
     const char *space;
+    Symbollist_types mode;
+    bool append;
 };
 
-struct arguments_s {
-    bool warning;
-    Caret_types caret;
-    bool quiet;
-    bool to_ascii;
+struct list_output_s {
+    const char *name;
     bool monitor;
     bool source;
     bool linenum;
+    bool verbose;
+    bool append;
+};
+
+struct cmdline_defines_s {
+    char *data;
+    size_t len;
+};
+
+struct include_list_s {
+    struct include_list_s *next;
+#if __STDC_VERSION__ >= 199901L
+    char path[];
+#elif __GNUC__ >= 3
+    char path[];
+#else
+    char path[1];
+#endif
+};
+
+struct arguments_s {
+    bool quiet;
+    bool to_ascii;
     bool longbranch;
     bool tasmcomp;
-    bool verbose;
     bool make_phony;
     uint8_t caseinsensitive;
     struct output_s *output;
@@ -67,9 +97,11 @@ struct arguments_s {
     const struct cpu_s *cpumode;
     struct symbol_output_s *symbol_output;
     size_t symbol_output_len;
-    const char *list;
+    struct include_list_s *include;
+    struct list_output_s list;
     const char *make;
-    const char *error;
+    struct cmdline_defines_s defines;
+    struct error_output_s error;
     unsigned int tab_size;
 };
 
@@ -99,7 +131,6 @@ struct diagnostics_s {
         bool variable;
     } unused;
     bool case_symbol;
-    bool switch_case;
     bool immediate;
     bool float_compare;
     bool leading_zeros;
@@ -112,9 +143,11 @@ struct diagnostics_s {
     bool page;
     bool macro_prefix;
     bool float_round;
+    bool size_larger;
 };
 
-extern int testarg(int *, char ***, struct file_s *);
+extern int testarg(int *, char ***);
+extern void destroy_arguments(void);
 extern struct arguments_s arguments;
 extern struct diagnostics_s diagnostics, diagnostic_errors;
 
