@@ -1,5 +1,5 @@
 /*
-    $Id: encoding.c 2175 2020-03-23 19:18:35Z soci $
+    $Id: encoding.c 2709 2021-09-18 18:40:01Z soci $
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -20,51 +20,23 @@
 #include "error.h"
 #include "string.h"
 #include "ternary.h"
-#include "unicode.h"
 #include "values.h"
 #include "64tass.h"
+#include "str.h"
 
-#include "strobj.h"
-#include "bytesobj.h"
-#include "typeobj.h"
-#include "errorobj.h"
+#include "encobj.h"
 
-struct encoding_s *actual_encoding;
-
-#define identmap (const uint8_t *)petscii_trans
-
-struct encoding_s {
-    str_t name;
-    str_t cfname;
-    bool empty;
-    bool failed;
-    ternary_tree escape;
-    size_t escape_length;
-    struct avltree trans;
-    struct avltree_node node;
-    uint8_t table[128];
-    uint32_t table_use[4];
-};
-
-struct trans2_s {
+struct translate_table_s {
     uint16_t start;
     uint8_t length;
     uint8_t offset;
 };
 
-struct escape_s {
-    size_t len;
-    uint8_t val[4];
-    uint8_t *data;
-};
-
-static struct avltree encoding_tree;
-
-static const struct trans2_s no_trans[] = {
+static const struct translate_table_s no_trans[] = {
     {0x0000, 255, 0x00},
 };
 
-static const struct trans2_s petscii_trans[] = {
+static const struct translate_table_s petscii_trans[] = {
     {0x0020,  32, 0x20}, /*  -@ */
     {0x0041,  25, 0xc1}, /* A-Z */
     {0x005b,   0, 0x5b}, /* [ */
@@ -145,7 +117,7 @@ static const struct trans2_s petscii_trans[] = {
 };
 
 /* PETSCII codes, must be sorted */
-static const char *petscii_esc =
+const char *const petscii_esc =
     "\x07" "{bell}\0"
     "\x90" "{black}\0"
     "\x90" "{blk}\0"
@@ -371,7 +343,7 @@ static const char *petscii_esc =
     "\x9e" "{yel}\0"
     "\x00" "\0";
 
-static const struct trans2_s petscii_screen_trans[] = {
+static const struct translate_table_s petscii_screen_trans[] = {
     {0x0020,  31, 0x20}, /*  -? */
     {0x0040,   0, 0x00}, /* @ */
     {0x0041,  25, 0x41}, /* A-Z */
@@ -453,7 +425,7 @@ static const struct trans2_s petscii_screen_trans[] = {
 };
 
 /* petscii screen codes, must be sorted */
-static const char *petscii_screen_esc =
+static const char petscii_screen_esc[] =
     "\x5f" "{cbm-*}\0"
     "\x66" "{cbm-+}\0"
     "\x5c" "{cbm--}\0"
@@ -545,7 +517,7 @@ static const char *petscii_screen_esc =
     "\x1e" "{up arrow}\0"
     "\x00" "\0";
 
-static const struct trans2_s no_screen_trans[] = {
+static const struct translate_table_s no_screen_trans[] = {
     {0x0000,  31, 0x80},
     {0x0020,  31, 0x20},
     {0x0040,  31, 0x00},
@@ -556,35 +528,22 @@ static const struct trans2_s no_screen_trans[] = {
     {0x00FF,   0, 0x5E},
 };
 
-static FAST_CALL int trans_compare(const struct avltree_node *aa, const struct avltree_node *bb)
-{
-    const struct trans_s *a = cavltree_container_of(aa, struct trans_s, node);
-    const struct trans_s *b = cavltree_container_of(bb, struct trans_s, node);
+struct encoding_s {
+    str_t name;
+    str_t cfname;
+    Enc *enc;
+    uint8_t pass;
+    struct avltree_node node;
+};
 
-    if (a->start > b->end) {
-        return -1;
-    }
-    if (a->end < b->start) {
-        return 1;
-    }
-    return 0;
-}
+static struct avltree encoding_tree;
 
 static FAST_CALL int encoding_compare(const struct avltree_node *aa, const struct avltree_node *bb)
 {
-    const struct encoding_s *a = cavltree_container_of(aa, struct encoding_s, node);
-    const struct encoding_s *b = cavltree_container_of(bb, struct encoding_s, node);
+    const str_t *a = &cavltree_container_of(aa, struct encoding_s, node)->cfname;
+    const str_t *b = &cavltree_container_of(bb, struct encoding_s, node)->cfname;
 
-    return str_cmp(&a->cfname, &b->cfname);
-}
-
-static void escape_free(void *e) {
-    struct escape_s *esc;
-    size_t i = (const uint8_t *)e - identmap;
-    if (i < 256) return;
-    esc = (struct escape_s *)e;
-    if (esc->data != esc->val) free(esc->data);
-    free(esc);
+    return str_cmp(a, b);
 }
 
 static void encoding_free(struct avltree_node *aa)
@@ -593,158 +552,58 @@ static void encoding_free(struct avltree_node *aa)
 
     free((char *)a->name.data);
     if (a->name.data != a->cfname.data) free((uint8_t *)a->cfname.data);
-    if (a->escape != NULL) ternary_cleanup(a->escape, escape_free);
+    val_destroy(Obj(a->enc));
     free(a);
 }
 
-static struct encoding_s *lasten = NULL;
-struct encoding_s *new_encoding(const str_t *name, linepos_t epoint)
+static Enc *enc_from_name(str_t *, linepos_t);
+
+static struct encoding_s *lasten;
+Enc *new_encoding(const str_t *name, linepos_t epoint)
 {
     struct avltree_node *b;
     struct encoding_s *tmp;
 
-    if (lasten == NULL) {
-        lasten = (struct encoding_s *)mallocx(sizeof *lasten);
-    }
+    if (lasten == NULL) new_instance(&lasten);
     str_cfcpy(&lasten->cfname, name);
     b = avltree_insert(&lasten->node, &encoding_tree, encoding_compare);
     if (b == NULL) { /* new encoding */
         str_cpy(&lasten->name, name);
         if (lasten->cfname.data == name->data) lasten->cfname = lasten->name;
         else str_cfcpy(&lasten->cfname, NULL);
-        lasten->escape = NULL;
-        lasten->escape_length = SIZE_MAX;
-        lasten->empty = true;
-        lasten->failed = false;
-        avltree_init(&lasten->trans);
-        memset(lasten->table_use, 0, sizeof(lasten->table_use));
         tmp = lasten;
         lasten = NULL;
-        return tmp;
+        tmp->enc = enc_from_name(&tmp->cfname, epoint);
+        tmp->pass = pass;
+    } else {
+        tmp = avltree_container_of(b, struct encoding_s, node);
+        if (tmp->pass != pass) {
+            tmp->enc->file_list = current_file_list;
+            tmp->enc->epoint = *epoint;
+        }
     }
-    tmp = avltree_container_of(b, struct encoding_s, node);
-    if (tmp->failed && tmp->empty) err_msg2(ERROR__EMPTY_ENCODI, NULL, epoint);
-    return tmp;            /* already exists */
+    return tmp->enc;
 }
 
-static struct transs_s {
-    struct trans_s transs[31];
-    struct transs_s *next;
-} *transs = NULL;
-
-static unsigned int transs_i = lenof(transs->transs);
-
-static struct trans_s *lasttr = NULL;
-struct trans_s *new_trans(struct trans_s *trans, struct encoding_s *enc, linepos_t epoint)
-{
-    struct avltree_node *b;
-    struct trans_s *tmp;
-    if (lasttr == NULL) {
-        if (transs_i == lenof(transs->transs)) {
-            struct transs_s *old = transs;
-            transs = (struct transs_s *)mallocx(sizeof *transs);
-            transs->next = old;
-            transs_i = 0;
-        }
-        lasttr = &transs->transs[transs_i++];
+static void add_trans(Enc *enc, const struct translate_table_s *table, size_t ln) {
+    size_t i;
+    struct linepos_s nopoint = {0, 0};
+    for (i = 0; i < ln; i++) {
+        struct character_range_s range;
+        uint32_t start = table[i].start;
+        if (start >= 0x8000) start += 0x10000;
+        range.start = start;
+        range.end = (start + table[i].length) & 0xffffff;
+        range.offset = table[i].offset;
+        enc_trans_add(enc, &range, &nopoint);
     }
-    lasttr->start = trans->start;
-    lasttr->end = trans->end;
-    lasttr->offset = trans->offset;
-    b = avltree_insert(&lasttr->node, &enc->trans, trans_compare);
-    if (b == NULL) { /* new encoding */
-        tmp = lasttr;
-        if (fixeddig && pass > max_pass) err_msg_cant_calculate(NULL, epoint);
-        fixeddig = false;
-        lasttr = NULL;
-        enc->empty = false;
-        if (trans->start < 128) memset(enc->table_use, 0, sizeof(enc->table_use));
-        return tmp;
-    }
-    return avltree_container_of(b, struct trans_s, node);            /* already exists */
 }
 
-bool new_escape(const str_t *v, Obj *val, struct encoding_s *enc, linepos_t epoint)
-{
-    struct escape_s **b2, *b, tmp;
-    Obj *val2;
-    struct iter_s iter;
-    uval_t uval;
-    size_t i, len;
-    uint8_t *d;
-    bool ret;
-
-    b2 = (struct escape_s **)ternary_insert(&enc->escape, v->data, v->data + v->len);
-    if (b2 == NULL) err_msg_out_of_memory();
-    b = *b2;
-    *b2 = NULL;
-
-    i = 0;
-    len = sizeof tmp.val;
-    d = tmp.val;
-
-    if (val->obj == STR_OBJ) {
-        Obj *tmp2 = bytes_from_str((Str *)val, epoint, BYTES_MODE_TEXT);
-        iter.data = tmp2; tmp2->obj->getiter(&iter);
-        val_destroy(tmp2);
-    } else { iter.data = val; val->obj->getiter(&iter); }
-
-    while ((val2 = iter.next(&iter)) != NULL) {
-        Error *err = val2->obj->uval(val2, &uval, 8, epoint);
-        if (err != NULL) {
-            err_msg_output_and_destroy(err);
-            uval = 0;
-        }
-        if (i >= len) {
-            if (i == sizeof tmp.val) {
-                len = 16;
-                d = (uint8_t *)mallocx(len);
-                memcpy(d, tmp.val, i);
-            } else {
-                len += 1024;
-                if (len < 1024) err_msg_out_of_memory(); /* overflow */
-                d = (uint8_t *)reallocx(d, len);
-            }
-        }
-        d[i++] = (uint8_t)uval;
-    }
-    iter_destroy(&iter);
-
-    if (b == NULL) { /* new escape */
-        if (i == 1) {
-            b = (struct escape_s *)(identmap + tmp.val[0]);
-        } else {
-            b = (struct escape_s *)mallocx(sizeof *b);
-            if (d == tmp.val) {
-                memcpy(b->val, tmp.val, i);
-                d = b->val;
-            } else if (i < len) {
-                d = (uint8_t *)reallocx(d, i);
-            }
-            b->len = i;
-            b->data = d;
-        }
-        *b2 = b;
-        if (v->len < enc->escape_length) enc->escape_length = v->len;
-        
-        enc->empty = false;
-        if (fixeddig && pass > max_pass) err_msg_cant_calculate(NULL, epoint);
-        fixeddig = false;
-        return false;
-    }
-    *b2 = b;
-    if (i == 1) return b != (struct escape_s *)(identmap + tmp.val[0]);
-    ret = (i != b->len || memcmp(d, b->data, i) != 0);
-    if (tmp.val != d) free(d);
-    return ret;            /* already exists */
-}
-
-static void add_esc(const char *s, struct encoding_s *enc) {
-    const uint8_t **b;
-    enc->empty = s[1] != 0;
+static void add_esc(Enc *enc, const char *s) {
+    enc->escape_char = '{';
     while (s[1] != 0) {
         size_t len = strlen(s + 1);
-        b = (const uint8_t **)ternary_insert(&enc->escape, (const uint8_t*)s + 1, (const uint8_t*)s + 1 + len);
+        const uint8_t **b = (const uint8_t **)ternary_insert(&enc->escapes, (const uint8_t*)s + 1, (const uint8_t*)s + 1 + len);
         if (b == NULL) err_msg_out_of_memory();
         *b = identmap + (uint8_t)s[0];
         if (enc->escape_length > len) enc->escape_length = len;
@@ -752,158 +611,43 @@ static void add_esc(const char *s, struct encoding_s *enc) {
     }
 }
 
-static void add_trans(const struct trans2_s *t, size_t ln, struct encoding_s *tmp) {
-    size_t i;
-    struct trans_s tmp2;
-    struct linepos_s nopoint = {0, 0};
-    for (i = 0; i < ln; i++) {
-        uint32_t start = t[i].start;
-        if (start >= 0x8000) start += 0x10000;
-        tmp2.start = start;
-        tmp2.end = start + t[i].length;
-        tmp2.offset = t[i].offset;
-        new_trans(&tmp2, tmp, &nopoint);
-    }
-}
+static bool ascii_mode;
 
-static struct {
-    size_t i, i2, j, len, len2;
-    bool err;
-    const uint8_t *data, *data2;
-    linepos_t epoint;
-    char mode;
-} encode_state;
-
-void encode_string_init(const Str *v, linepos_t epoint) {
-    encode_state.i = 0;
-    encode_state.j = 0;
-    encode_state.len2 = 0;
-    encode_state.len = v->len;
-    encode_state.data = v->data;
-    encode_state.epoint = epoint;
-    encode_state.err = false;
-}
-
-void encode_error(Error_types no) {
-    if (!encode_state.err) {
-        struct linepos_s epoint = *encode_state.epoint;
-        epoint.pos = interstring_position(&epoint, encode_state.data, encode_state.i2);
-        err_msg2(no, NULL, &epoint);
-        encode_state.err = true;
-    }
-}
-
-int encode_string(void) {
-    uchar_t ch;
-    unsigned int ln;
-    const struct avltree_node *c;
-    const struct trans_s *t;
-    struct trans_s tmp;
-
-    if (encode_state.j < encode_state.len2) {
-        return encode_state.data2[encode_state.j++];
-    }
-next:
-    if (encode_state.i >= encode_state.len) return EOF;
-    encode_state.i2 = encode_state.i;
-    if (encode_state.len - encode_state.i >= actual_encoding->escape_length) {
-        size_t len = encode_state.len - encode_state.i;
-        const struct escape_s *e = (struct escape_s *)ternary_search(actual_encoding->escape, encode_state.data + encode_state.i, &len);
-        if (e != NULL) {
-            size_t i;
-            encode_state.i += len;
-            i = (const uint8_t *)e - identmap;
-            if (i < 256) {
-                encode_state.len2 = 1;
-                encode_state.j = 1;
-                return i;
-            }
-            encode_state.data2 = e->data;
-            encode_state.len2 = e->len;
-            if (encode_state.len2 < 1) goto next;
-            encode_state.j = 1;
-            return e->data[0];
+static Enc *enc_from_name(str_t *name, linepos_t epoint) {
+    Enc *enc = Enc(new_enc(current_file_list, epoint));
+    bool oldfixeddig = fixeddig;
+    uint8_t oldpass = pass;
+    fixeddig = false;
+    pass = 255;
+    if (name->len == 4 && memcmp(name->data, "none", 4) == 0) {
+        if (!ascii_mode) {
+            add_trans(enc, no_trans, lenof(no_trans));
+        } else {
+            add_esc(enc, petscii_esc);
+            add_trans(enc, petscii_trans, lenof(petscii_trans));
+        }
+    } else if (name->len == 6 && memcmp(name->data, "screen", 6) == 0) {
+        if (!ascii_mode) {
+            add_trans(enc, no_screen_trans, lenof(no_screen_trans));
+        } else {
+            add_esc(enc, petscii_screen_esc);
+            add_trans(enc, petscii_screen_trans, lenof(petscii_screen_trans));
         }
     }
-    ch = encode_state.data[encode_state.i];
-    if ((ch & 0x80) != 0) ln = utf8in(encode_state.data + encode_state.i, &ch); else {
-        if ((actual_encoding->table_use[ch / 32] & (1 << (ch % 32))) != 0) {
-            encode_state.i++;
-            return actual_encoding->table[ch];
-        }
-        ln = 1;
-    }
-    tmp.start = tmp.end = ch;
-
-    c = avltree_lookup(&tmp.node, &actual_encoding->trans, trans_compare);
-    if (c != NULL) {
-        t = cavltree_container_of(c, struct trans_s, node);
-        if (tmp.start >= t->start && tmp.end <= t->end) {
-            encode_state.i += ln;
-            if (ch < 0x80) {
-                actual_encoding->table_use[ch / 32] |= 1 << (ch % 32);
-                actual_encoding->table[ch] = (uint8_t)(ch - t->start + t->offset);
-            }
-            return (uint8_t)(ch - t->start + t->offset);
-        }
-    }
-    if (!encode_state.err && (!actual_encoding->empty || !actual_encoding->failed)) {
-        struct linepos_s epoint = *encode_state.epoint;
-        epoint.pos = interstring_position(&epoint, encode_state.data, encode_state.i);
-        err_msg_unknown_char(ch, &actual_encoding->name, &epoint);
-        actual_encoding->failed = true;
-        encode_state.err = true;
-    }
-    encode_state.i += ln;
-    return 256 + '?';
+    pass = oldpass;
+    fixeddig = oldfixeddig;
+    return enc;
 }
 
 void init_encoding(bool toascii)
 {
-    struct encoding_s *tmp;
-    static const str_t none_enc = {(const uint8_t *)"none", 4};
-    static const str_t screen_enc = {(const uint8_t *)"screen", 6};
-    struct linepos_s nopoint = {0, 0};
-
     avltree_init(&encoding_tree);
-
-    if (!toascii) {
-        tmp = new_encoding(&none_enc, &nopoint);
-        if (tmp == NULL) {
-            return;
-        }
-        add_trans(no_trans, lenof(no_trans), tmp);
-
-        tmp = new_encoding(&screen_enc, &nopoint);
-        if (tmp == NULL) {
-            return;
-        }
-        add_trans(no_screen_trans, lenof(no_screen_trans), tmp);
-    } else {
-        tmp = new_encoding(&none_enc, &nopoint);
-        if (tmp == NULL) {
-            return;
-        }
-        add_esc(petscii_esc, tmp);
-        add_trans(petscii_trans, lenof(petscii_trans), tmp);
-
-        tmp = new_encoding(&screen_enc, &nopoint);
-        if (tmp == NULL) {
-            return;
-        }
-        add_esc(petscii_screen_esc, tmp);
-        add_trans(petscii_screen_trans, lenof(petscii_screen_trans), tmp);
-    }
+    ascii_mode = toascii;
+    lasten = NULL;
 }
 
 void destroy_encoding(void)
 {
     avltree_destroy(&encoding_tree, encoding_free);
     free(lasten);
-
-    while (transs != NULL) {
-        struct transs_s *old = transs;
-        transs = transs->next;
-        free(old);
-    }
 }
