@@ -1,5 +1,5 @@
 /*
-    $Id: encoding.h 1816 2019-01-13 18:35:30Z soci $
+    $Id: encoding.h 2828 2022-10-20 20:29:55Z soci $
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -18,32 +18,16 @@
 */
 #ifndef ENCODING_H
 #define ENCODING_H
-#include "avl.h"
 #include "stdbool.h"
-#include "inttypes.h"
-#include "errors_e.h"
-
-struct encoding_s;
-struct Obj;
-
-struct trans_s {
-    uint32_t start;
-    uint32_t end : 24;
-    uint32_t offset : 8;
-    struct avltree_node node;
-};
-
-extern struct encoding_s *actual_encoding;
 
+struct Enc;
 struct str_t;
-struct Str;
+struct linepos_s;
 
-extern struct encoding_s *new_encoding(const struct str_t *, linepos_t);
-extern struct trans_s *new_trans(struct trans_s *, struct encoding_s *, linepos_t);
-extern bool new_escape(const struct str_t *, struct Obj *, struct encoding_s *, linepos_t);
-extern void encode_string_init(const struct Str *, linepos_t);
-extern int encode_string(void);
-extern void encode_error(Error_types);
+extern const char *const petscii_esc;
+#define identmap (const uint8_t *)petscii_esc
+
+extern struct Enc *new_encoding(const struct str_t *, const struct linepos_s *);
 extern void init_encoding(bool);
 extern void destroy_encoding(void);
 #endif
