@@ -1,5 +1,5 @@
 /*
-    $Id: errors_e.h 2079 2019-11-11 20:40:59Z soci $
+    $Id: errors_e.h 2887 2022-11-01 09:07:24Z soci $
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -30,13 +30,12 @@ typedef enum Error_types {
     ERROR_______OLD_NEQ,
     ERROR______OLD_GOTO,
     ERROR____OLD_STRING,
-    ERROR_______OLD_ENC,
     ERROR__CONST_RESULT,
     ERROR____IND_RESULT,
 #ifdef _WIN32
     ERROR___INSENSITIVE,
 #endif
-#if defined _WIN32 || defined __WIN32__ || defined __EMX__ || defined __MSDOS__ || defined __DOS__
+#if defined _WIN32 || defined __WIN32__ || defined __MSDOS__ || defined __DOS__
     ERROR_____BACKSLASH,
 #else
     ERROR__RESERVED_CHR,
@@ -47,7 +46,6 @@ typedef enum Error_types {
     ERROR______SIMPLIFY,
     ERROR_____REDUNDANT,
     ERROR_____OLD_EQUAL,
-    ERROR_DUPLICATECASE,
     ERROR_NONIMMEDCONST,
     ERROR_FLOAT_COMPARE,
     ERROR_LEADING_ZEROS,
@@ -71,6 +69,7 @@ typedef enum Error_types {
     ERROR_CANT_CROSS_BA,
     ERROR_OUTOF_SECTION,
     ERROR_NEGFRAC_POWER,
+    ERROR_ZERO_NEGPOWER,
     ERROR_SQUARE_ROOT_N,
     ERROR_LOG_NON_POSIT,
     ERROR___MATH_DOMAIN,
@@ -87,6 +86,8 @@ typedef enum Error_types {
     ERROR__NOT_KEYVALUE,
     ERROR_____CANT_IVAL,
     ERROR_____CANT_UVAL,
+    ERROR____CANT_IVAL2,
+    ERROR____CANT_UVAL2,
     ERROR______NOT_UVAL,
     ERROR_CANT_BROADCAS,
     ERROR_____CANT_SIGN,
@@ -104,10 +105,10 @@ typedef enum Error_types {
     ERROR_____NOT_BANK0,
     ERROR_OUT_OF_MEMORY,
     ERROR__ADDR_COMPLEX,
-    ERROR__EMPTY_ENCODI,
     ERROR_MISSING_CLOSE,
     ERROR__MISSING_OPEN,
     ERROR__MISSING_LOOP,
+    ERROR_NEGATIVE_SIZE,
     ERROR__USER_DEFINED,
     ERROR_NO_ADDRESSING,
     ERROR___NO_REGISTER,
@@ -116,7 +117,6 @@ typedef enum Error_types {
     ERROR_BRANCH_TOOFAR,
     ERROR____PTEXT_LONG,
     ERROR______EXPECTED,
-    ERROR___NOT_ALLOWED,
     ERROR_RESERVED_LABL,
     ERROR___UNKNOWN_CPU,
 
@@ -124,6 +124,7 @@ typedef enum Error_types {
     ERROR__INVALID_OPER,
     ERROR__INVALID_CONV,
     ERROR__WRONG_ARGNUM,
+    ERROR____WRONG_TYPE,
     ERROR____STILL_NONE,
 
     ERROR_CANT_FINDFILE = 0xc0,
@@ -133,6 +134,7 @@ typedef enum Error_types {
     ERROR_CANT_WRTE_LBL,
     ERROR_CANT_WRTE_MAK,
     ERROR_CANT_WRTE_ERR,
+    ERROR_CANT_WRTE_MAP,
     ERROR_FILERECURSION,
     ERROR__MACRECURSION,
     ERROR__FUNRECURSION,
