@@ -1,5 +1,5 @@
 /*
-    $Id: eval.h 2061 2019-11-02 19:13:11Z soci $
+    $Id: eval.h 2737 2021-10-06 20:50:52Z soci $
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -21,14 +21,15 @@
 #include "attributes.h"
 #include "inttypes.h"
 #include "stdbool.h"
+#include "oper_e.h"
 
-struct List;
 struct Label;
 struct Oper;
 struct Obj;
+struct Funcargs;
 
 struct oper_s {
-    struct Oper *op;
+    Oper_types op;
     struct Obj *v1;
     struct Obj *v2;
     linepos_t epoint;
@@ -38,10 +39,10 @@ struct oper_s {
 };
 typedef struct oper_s *oper_t;
 
-extern bool get_exp(int, unsigned int, unsigned int, linepos_t);
+extern bool get_exp(int, argcount_t, argcount_t, linepos_t);
 extern struct values_s *get_val(void);
-extern struct Obj *pull_val(struct linepos_s *);
-extern size_t get_val_remaining(void);
+extern struct Obj *pull_val(void);
+extern argcount_t get_val_remaining(void);
 extern void destroy_eval(void);
 extern void init_eval(void);
 extern void eval_enter(void);
@@ -49,11 +50,10 @@ extern void eval_leave(void);
 extern FAST_CALL size_t get_label(const uint8_t *);
 extern MUST_CHECK struct Obj *get_star_value(address_t, struct Obj *);
 extern MUST_CHECK struct Obj *get_star(void);
-extern struct Obj *get_vals_tuple(void);
-extern struct Obj *get_vals_addrlist(struct linepos_s *);
-extern MUST_CHECK struct Error *indexoffs(struct Obj *, size_t, size_t *, linepos_t);
-extern MUST_CHECK struct Obj *sliceparams(const struct List *, size_t, uval_t *, ival_t *, ival_t *, ival_t *, linepos_t);
+extern MUST_CHECK struct Obj *get_vals_tuple(void);
+extern void get_vals_funcargs(struct Funcargs *f);
 extern void touch_label(struct Label *);
+extern MUST_CHECK struct Obj *calc2_lxor(oper_t, bool);
 
 struct values_s {
     struct Obj *val;
