@@ -1,5 +1,5 @@
 /*
-    $Id: file.h 2125 2019-12-23 19:50:58Z soci $
+    $Id: file.h 2833 2022-10-22 10:01:11Z soci $
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -19,57 +19,60 @@
 #ifndef FILE_H
 #define FILE_H
 #include <stdio.h>
-#include "attributes.h"
 #include "stdbool.h"
 #include "inttypes.h"
-#include "avl.h"
-#include "str.h"
 
 typedef enum Encoding_types {
-    E_UNKNOWN, E_UTF8, E_UTF16LE, E_UTF16BE, E_ISO
+    E_UNKNOWN, E_UTF8, E_UTF16LE, E_UTF16BE, E_ISO, E_RAW
 } Encoding_types;
 
+typedef enum File_open_type {
+    FILE_OPEN_STDIN, FILE_OPEN_BINARY, FILE_OPEN_DEFINES, FILE_OPEN_SOURCE
+} File_open_type;
+
+typedef uint32_t filesize_t;
+
+struct file_data_s {
+    uint8_t *data;
+    filesize_t len;
+    bool read;
+};
+
 struct file_s {
     const char *name;
-    const char *realname;
-    str_t base;
     int hash;
     uint8_t *nomacro;
-    size_t *line;
-    line_t lines;
-    uint8_t *data;    /* data */
-    size_t len;       /* length */
-    uint16_t open;    /* open/not open */
-    uint16_t uid;     /* uid */
-    int type;
+    filesize_t *line;
+    linenum_t lines;
+    struct file_data_s source;
+    struct file_data_s binary;
     int err_no;
+    bool open;
     bool read_error;
     bool portable;
+    bool cmdline;
     uint8_t pass;
     uint8_t entercount;
+    uint16_t uid;
     Encoding_types encoding;
 };
 
+#define not_in_file(a, b) ((size_t)((a) - (1 ? (b) : (struct file_s *)(void *)(b))->source.data) >= (b)->source.len)
+
 struct star_s {
-    line_t line;
+    linenum_t line, vline;
     address_t addr;
-    struct avltree tree;
-    struct avltree_node node;
+    uint8_t pass;
 };
 
-static inline bool dash_name(const char *name) {
-    return (name[0] == '-' && name[1] == 0);
-}
+struct str_t;
+struct file_list_s;
 
-extern struct file_s *openfile(const char *, const char *, int, const struct str_t *, linepos_t);
-extern void closefile(struct file_s*);
-extern struct star_s *new_star(line_t, bool *);
+extern struct file_s *file_open(const struct str_t *, const struct file_list_s *, File_open_type, linepos_t);
+extern struct star_s *new_star(linenum_t);
+extern struct star_s *init_star(linenum_t);
 extern void destroy_file(void);
 extern void init_file(void);
-extern FILE *file_open(const char *, const char *);
-extern void include_list_add(const char *);
-extern size_t get_base(const char *);
-extern char *get_path(const struct str_t *, const char *);
 extern void makefile(int, char *[], bool);
 
 #endif
