@@ -1,5 +1,5 @@
 /*
-    $Id: listing.c 2098 2019-11-24 05:07:52Z soci $
+    $Id: listing.c 2924 2022-12-22 08:52:34Z soci $
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -31,6 +31,8 @@
 #include "obj.h"
 #include "values.h"
 #include "arguments.h"
+#include "macro.h"
+#include "version.h"
 
 #define LINE_WIDTH 8
 #define ADDR_WIDTH 8
@@ -44,13 +46,13 @@ const uint8_t *llist = NULL;
 
 typedef struct Listing {
     size_t c;
-    unsigned int i;
-    char s[LINE_WIDTH + ADDR_WIDTH + LADDR_WIDTH + 3*16];
+    char *s;
+    char buf[LINE_WIDTH + ADDR_WIDTH + LADDR_WIDTH + 3*16];
     char hex[16];
     const uint16_t *disasm;
     const uint32_t *mnemonic;
     struct {
-        size_t addr, laddr, hex, monitor, source;
+        unsigned int addr, laddr, hex, monitor, source;
     } columns;
     FILE *flist;
     uint16_t lastfile;
@@ -60,8 +62,8 @@ typedef struct Listing {
 } Listing;
 
 static void flushbuf(Listing *ls) {
-    ls->c += fwrite(ls->s, 1, ls->i, ls->flist);
-    ls->i = 0;
+    ls->c += fwrite(ls->buf, 1, (size_t)(ls->s - ls->buf), ls->flist);
+    ls->s = ls->buf;
 }
 
 static void newline(Listing *ls) {
@@ -69,24 +71,27 @@ static void newline(Listing *ls) {
     ls->c = 0;
 }
 
-static void padding2(Listing *ls, size_t t) {
-    if (ls->c + ls->i >= t) {ls->s[ls->i++] = '\n'; flushbuf(ls); ls->c = 0;}
-    ls->c += ls->i;
-    if (ls->tab_size > 1) {
-        ls->c -= ls->c % ls->tab_size;
-        while (ls->c + ls->tab_size <= t) { ls->c += ls->tab_size; ls->s[ls->i++] = '\t'; }
+static void padding2(Listing *ls, unsigned int t) {
+    unsigned int ts;
+    size_t c = ls->c + (size_t)(ls->s - ls->buf);
+    if (c >= t) {*ls->s++ = '\n'; flushbuf(ls); c = 0;}
+    ts = ls->tab_size;
+    if (ts > 1) {
+        c -= c % ts;
+        while (c + ts <= t) { c += ts; *ls->s++ = '\t'; }
     }
-    while (ls->c < t) { ls->c++; ls->s[ls->i++] = ' '; }
-    ls->c -= ls->i;
+    while (c < t) { c++; *ls->s++ = ' '; }
+    ls->c = c - (size_t)(ls->s - ls->buf);
 }
 
 static inline void out_hex(Listing *ls, unsigned int c) {
-    ls->s[ls->i++] = ls->hex[(c >> 4) & 15];
-    ls->s[ls->i++] = ls->hex[c & 15];
+    ls->s[0] = ls->hex[(c >> 4) & 15];
+    ls->s[1] = ls->hex[c & 15];
+    ls->s += 2;
 }
 
 static void out_byte(Listing *ls, unsigned int adr) {
-    ls->s[ls->i++] = '$';
+    *ls->s++ = '$';
     out_hex(ls, adr);
 }
 
@@ -107,65 +112,71 @@ static void out_zp(Listing *ls, unsigned int adr) {
 }
 
 static void out_db(Listing *ls, unsigned int adr) {
-    ls->s[ls->i++] = '$';
+    *ls->s++ = '$';
     if (databank != 0) out_hex(ls, databank);
     out_hex(ls, adr >> 8);
     out_hex(ls, adr);
 }
 
 static void out_pb(Listing *ls, unsigned int adr) {
-    ls->s[ls->i++] = '$';
-    if (current_address->l_address.bank != 0) out_hex(ls, current_address->l_address.bank >> 16);
+    *ls->s++ = '$';
+    if (current_address->l_address > 0xffff) out_hex(ls, current_address->l_address >> 16);
     out_hex(ls, adr >> 8);
     out_hex(ls, adr);
 }
 
 static void out_bit(Listing *ls, unsigned int cod, unsigned int c) {
-    ls->s[ls->i++] = 0x30 + ((cod >> 4) & 7);
-    ls->s[ls->i++] = ',';
+    ls->s[0] = (char)('0' + ((cod >> 4) & 7));
+    ls->s[1] = ',';
+    ls->s += 2;
     out_zp(ls, c);
 }
 
 static void out_txt(Listing *ls, const char *s) {
     while (*s != '\0') {
-        ls->s[ls->i++] = *s++;
+        *ls->s++ = *s++;
     }
 }
 
-MUST_CHECK Listing *listing_open(const char *filename, int argc, char *argv[]) {
+static Listing *listing;
+
+void listing_open(const struct list_output_s *output, int argc, char *argv[]) {
+    static Listing listing2;
     Listing *ls;
-    struct linepos_s nopoint = {0, 0};
     time_t t;
     int i;
     FILE *flist;
 
-    flist = dash_name(filename) ? stdout : file_open(filename, "wt");
+    flist = dash_name(output->name) ? stdout : fopen_utf8(output->name, output->append ? "at" : "wt");
     if (flist == NULL) {
-        err_msg_file(ERROR_CANT_WRTE_LST, filename, &nopoint);
-        return NULL;
+        err_msg_file2(ERROR_CANT_WRTE_LST, output->name);
+        listing = NULL;
+        return;
     }
+    if (flist == stdout && fflush(flist) != 0) setvbuf(flist, NULL, _IOLBF, 1024);
     clearerr(flist); errno = 0;
 
-    ls = (Listing *)mallocx(sizeof *ls);
+    ls = &listing2;
 
     memcpy(ls->hex, "0123456789abcdef", 16);
-    ls->filename = filename;
+    ls->filename = output->name;
     ls->flist = flist;
-    ls->linenum = arguments.linenum;
+    ls->linenum = arguments.list.linenum;
     ls->pccolumn = listing_pccolumn;
-    ls->columns.addr = arguments.linenum ? LINE_WIDTH : 0;
+    ls->columns.addr = arguments.list.linenum ? LINE_WIDTH : 0;
     ls->columns.laddr = ls->columns.addr + ADDR_WIDTH;
     ls->columns.hex = ls->columns.laddr + (ls->pccolumn ? LADDR_WIDTH : 0);
     ls->columns.monitor = ls->columns.hex + HEX_WIDTH;
-    ls->columns.source = ls->columns.monitor + (arguments.monitor ? MONITOR_WIDTH : 0);
+    ls->columns.source = ls->columns.monitor + (arguments.list.monitor ? MONITOR_WIDTH : 0);
     ls->tab_size = arguments.tab_size;
-    ls->verbose = arguments.verbose;
-    ls->monitor = arguments.monitor;
-    ls->source = arguments.source;
+    ls->verbose = arguments.list.verbose;
+    ls->monitor = arguments.list.monitor;
+    ls->source = arguments.list.source;
     ls->lastfile = 0;
-    ls->i = 0;
+    ls->s = ls->buf;
 
-    fputs("\n; 64tass Turbo Assembler Macro V" VERSION " listing file\n;", flist);
+    if (!output->append) fputs("\n; 64tass Turbo Assembler Macro V" VERSION " listing file", flist);
+    fputs("\n;", flist);
     for (i = 0; i < argc; i++) {
         putc(' ', flist);
         argv_print(argv[i], flist);
@@ -194,19 +205,19 @@ MUST_CHECK Listing *listing_open(const char *filename, int argc, char *argv[]) {
     }
     flushbuf(ls);
     newline(ls);
-    return ls;
+    listing = ls;
 }
 
-void listing_close(Listing *ls) {
-    struct linepos_s nopoint = {0, 0};
+void listing_close(void) {
+    Listing *const ls = listing;
     int err;
     if (ls == NULL) return;
 
     fputs("\n;******  End of listing\n", ls->flist);
     err = ferror(ls->flist);
     err |= (ls->flist != stdout) ? fclose(ls->flist) : fflush(ls->flist);
-    if (err != 0 && errno != 0) err_msg_file(ERROR_CANT_WRTE_LST, ls->filename, &nopoint);
-    free(ls);
+    if (err != 0 && errno != 0) err_msg_file2(ERROR_CANT_WRTE_LST, ls->filename);
+    listing = NULL;
 }
 
 static bool printllist(Listing *ls) {
@@ -232,14 +243,14 @@ static void printdec(Listing *ls, uint32_t dec) {
     for (; i < 9; i++) {
         uint32_t a = dec / d[i];
         dec = dec % d[i];
-        ls->s[ls->i++] = '0' + a;
+        *ls->s++ = (char)('0' + a);
     }
-    ls->s[ls->i++] = '0' + dec;
+    *ls->s++ = (char)('0' + dec);
 }
 
 static void printfile(Listing *ls) {
-    ls->s[ls->i++] = ':';
-    printdec(ls, ls->lastfile - 1);
+    *ls->s++ = ':';
+    printdec(ls, ls->lastfile - 1U);
 }
 
 static void printline(Listing *ls) {
@@ -251,40 +262,42 @@ static void printline(Listing *ls) {
     printfile(ls);
 }
 
-FAST_CALL void listing_equal(Listing *ls, Obj *val) {
+FAST_CALL void listing_equal(Obj *val) {
+    Listing *const ls = listing;
     if (ls == NULL) return;
-    if (nolisting != 0 || !ls->source || temporary_label_branch != 0) return;
+    if (nolisting != 0 || !ls->source || in_function) return;
     if (ls->linenum) {
         printline(ls);
         padding2(ls, ls->columns.addr);
         flushbuf(ls);
     }
     putc('=', ls->flist);
-    ls->c += val_print(val, ls->flist) + 1;
+    ls->c += val_print(val, ls->flist, ls->verbose ? SIZE_MAX : ls->columns.source - 2) + 1;
     printllist(ls);
     newline(ls);
 }
 
-static void printaddr(Listing *ls, char pre, address_t addr, address_t addr2) {
-    ls->s[ls->i++] = pre;
-    for (;;) {
-        if (addr > 0xffff) {
-            if (addr >> 24 != 0) out_hex(ls, addr >> 24);
-            out_hex(ls, addr >> 16);
-        }
-        out_hex(ls, addr >> 8);
-        out_hex(ls, addr);
-        if (addr2 == addr || !ls->pccolumn) return;
-        addr = addr2;
-        padding2(ls, ls->columns.laddr);
+static void printaddr(Listing *ls, address_t addr) {
+    if (addr > 0xffff) {
+        if (addr >> 24 != 0) out_hex(ls, addr >> 24);
+        out_hex(ls, addr >> 16);
     }
+    out_hex(ls, addr >> 8);
+    out_hex(ls, addr);
+}
+
+static void printaddr2(Listing *ls, address_t addr, address_t addr2) {
+    printaddr(ls, addr);
+    if (addr2 == addr || !ls->pccolumn) return;
+    padding2(ls, ls->columns.laddr);
+    printaddr(ls, addr2);
 }
 
 static void printhex(Listing *ls, unsigned int cod, uint32_t adr, int ln) {
     padding2(ls, ls->columns.hex);
     out_hex(ls, cod);
     while (ln > 0) {
-        ls->s[ls->i++] = ' ';
+        *ls->s++ = ' ';
         out_hex(ls, adr);
         adr >>= 8;
         ln--;
@@ -297,7 +310,7 @@ static void printhex2(Listing *ls, unsigned int ln, const uint8_t *data) {
     for (;;) {
         out_hex(ls, data[i++]);
         if (i >= ln) break;
-        ls->s[ls->i++] = ' ';
+        *ls->s++ = ' ';
     }
 }
 
@@ -308,21 +321,22 @@ static void printmon(Listing *ls, unsigned int cod, int ln, uint32_t adr) {
 
     padding2(ls, ls->columns.monitor);
     mnem = ls->mnemonic[ls->disasm[cod] & 0xff];
-    ls->s[ls->i++] = (char)(mnem >> 16);
-    ls->s[ls->i++] = (char)(mnem >> 8);
-    ls->s[ls->i++] = (char)(mnem);
-    ls->s[ls->i++] = ' ';
+    ls->s[0] = (char)(mnem >> 16);
+    ls->s[1] = (char)(mnem >> 8);
+    ls->s[2] = (char)(mnem);
+    ls->s[3] = ' ';
+    ls->s += 4;
 
     type = (Adr_types)(ls->disasm[cod] >> 8);
     mode = addr_modes[type];
-    if (*mode != ' ') ls->s[ls->i++] = *mode;
+    if (*mode != ' ') *ls->s++ = *mode;
     mode++;
     switch (type) {
-    case ADR_IMPLIED: ls->i--; return;
+    case ADR_IMPLIED: ls->s--; return;
     case ADR_REG: return;
     case ADR_IMMEDIATE:
         switch (ln) {
-        default: ls->i -= 2; return;
+        default: ls->s -= 2; return;
         case 0: return;
         case 1: out_byte(ls, adr); return;
         case 2: out_word(ls, adr); return;
@@ -349,35 +363,36 @@ static void printmon(Listing *ls, unsigned int cod, int ln, uint32_t adr) {
     case ADR_ZP_Y: out_zp(ls, adr); break;
     case ADR_ADDR_I:
     case ADR_ADDR_LI: out_word(ls, adr); break;
-    case ADR_REL: if (ln > 0) out_pb(ls, (address_t)((int8_t)adr + (int)current_address->l_address.address)); else ls->i--; return;
+    case ADR_REL: if (ln > 0) out_pb(ls, (address_t)((int8_t)adr + (int)current_address->l_address)); else ls->s--; return;
     case ADR_BIT_ZP_REL:
         out_bit(ls, cod, adr);
-        ls->s[ls->i++] = ',';
-        out_pb(ls, (address_t)((int8_t)(adr >> 8) + (int)current_address->l_address.address));
+        *ls->s++ = ',';
+        out_pb(ls, (address_t)((int8_t)(adr >> 8) + (int)current_address->l_address));
         return;
-    case ADR_REL_L: if (ln > 0) out_pb(ls, adr + (((cod & 0x0F) == 3) ? -1u : 0) + current_address->l_address.address); else ls->i--; return;
-    case ADR_MOVE: out_byte(ls, adr >> 8); ls->s[ls->i++] = ','; ls->s[ls->i++] = '#'; out_byte(ls, adr); return;
+    case ADR_REL_L: if (ln > 0) out_pb(ls, adr + (((cod & 0x0F) == 3) ? -1U : 0) + current_address->l_address); else ls->s--; return;
+    case ADR_MOVE: out_byte(ls, adr >> 8); ls->s[0] = ','; ls->s[1] = '#'; ls->s += 2; out_byte(ls, adr); return;
     }
-    while (*mode != 0) ls->s[ls->i++] = *mode++;
+    while (*mode != 0) *ls->s++ = *mode++;
 }
 
 static void printsource(Listing *ls, linecpos_t pos) {
-    while (llist[pos-1] == 0x20 || llist[pos-1] == 0x09) pos--;
+    while (pos > 0 && (llist[pos-1] == 0x20 || llist[pos-1] == 0x09)) pos--;
     padding2(ls, ls->columns.source);
     flushbuf(ls);
     printable_print2(llist, ls->flist, pos);
     newline(ls);
 }
 
-FAST_CALL void listing_equal2(Listing *ls, Obj *val, linecpos_t pos) {
+FAST_CALL void listing_equal2(Obj *val, linecpos_t pos) {
+    Listing *const ls = listing;
     if (ls == NULL) return;
-    if (nolisting != 0 || !ls->source || temporary_label_branch != 0) return;
+    if (nolisting != 0 || !ls->source || in_function) return;
     if (ls->linenum) {
         printline(ls);
         padding2(ls, ls->columns.addr);
     }
     putc('=', ls->flist);
-    ls->c += val_print(val, ls->flist) + 1;
+    ls->c += val_print(val, ls->flist, ls->verbose ? SIZE_MAX : ls->columns.source - 2) + 1;
     if (ls->verbose) {
         printllist(ls);
         newline(ls);
@@ -388,13 +403,14 @@ FAST_CALL void listing_equal2(Listing *ls, Obj *val, linecpos_t pos) {
 }
 
 
-FAST_CALL void listing_line(Listing *ls, linecpos_t pos) {
+FAST_CALL void listing_line(linecpos_t pos) {
+    Listing *const ls = listing;
     size_t i;
-    if (nolisting != 0  || temporary_label_branch != 0 || llist == NULL) return;
+    if (nolisting != 0  || in_function || llist == NULL) return;
     if (ls == NULL) {
         address_t addr;
-        if (!fixeddig || constcreated || listing_pccolumn || !arguments.source) return;
-        addr = (current_address->l_address.address & 0xffff) | current_address->l_address.bank;
+        if (!fixeddig || constcreated || listing_pccolumn || !arguments.list.source) return;
+        addr = current_address->l_address;
         i = 0;
         while (i < pos && (llist[i] == 0x20 || llist[i] == 0x09)) i++;
         if (i < pos && current_address->address != addr) listing_pccolumn = true;
@@ -408,10 +424,11 @@ FAST_CALL void listing_line(Listing *ls, linecpos_t pos) {
             printline(ls);
             padding2(ls, ls->columns.addr);
         }
-        printaddr(ls, '.', current_address->address, (current_address->l_address.address & 0xffff) | current_address->l_address.bank);
+        *ls->s++ = '.';
+        printaddr2(ls, current_address->address, current_address->l_address);
     }
     if (ls->verbose) {
-        if (ls->i != 0) flushbuf(ls);
+        if (ls->s != ls->buf) flushbuf(ls);
         if (llist[i] != 0) {
             if (ls->c == 0 && ls->linenum) printline(ls);
             padding2(ls, ls->columns.source);
@@ -420,22 +437,20 @@ FAST_CALL void listing_line(Listing *ls, linecpos_t pos) {
         }
         newline(ls);
     } else {
-        if (ls->c + ls->i != 0) printsource(ls, pos);
-        else if (ls->i != 0) flushbuf(ls);
+        if (ls->c != 0 || ls->s != ls->buf) printsource(ls, pos);
     }
     llist = NULL;
 }
 
-FAST_CALL void listing_line_cut(Listing *ls, linecpos_t pos) {
+FAST_CALL void listing_line_cut(linecpos_t pos) {
+    Listing *const ls = listing;
     size_t i;
-    if (nolisting != 0 || temporary_label_branch != 0 || llist == NULL) return;
+    if (nolisting != 0 || in_function || llist == NULL) return;
     if (ls == NULL) {
-        address_t addr;
-        if (!fixeddig || constcreated || listing_pccolumn || !arguments.source) return;
-        addr = (current_address->l_address.address & 0xffff) | current_address->l_address.bank;
+        if (!fixeddig || constcreated || listing_pccolumn || !arguments.list.source) return;
         i = 0;
         while (i < pos && (llist[i] == 0x20 || llist[i] == 0x09)) i++;
-        if (i < pos && current_address->address != addr) listing_pccolumn = true;
+        if (i < pos && current_address->address != current_address->l_address) listing_pccolumn = true;
         return;
     }
     if (!ls->source) return;
@@ -446,15 +461,17 @@ FAST_CALL void listing_line_cut(Listing *ls, linecpos_t pos) {
             printline(ls);
             padding2(ls, ls->columns.addr);
         }
-        printaddr(ls, '.', current_address->address, (current_address->l_address.address & 0xffff) | current_address->l_address.bank);
+        *ls->s++ = '.';
+        printaddr2(ls, current_address->address, current_address->l_address);
         printsource(ls, pos);
     }
     llist = NULL;
 }
 
-FAST_CALL void listing_line_cut2(Listing *ls, linecpos_t pos) {
+FAST_CALL void listing_line_cut2(linecpos_t pos) {
+    Listing *const ls = listing;
     if (ls == NULL || !ls->verbose || llist == NULL) return;
-    if (nolisting == 0 && ls->source && temporary_label_branch == 0) {
+    if (nolisting == 0 && ls->source && !in_function) {
         if (ls->linenum) printline(ls);
         padding2(ls, ls->columns.source);
         flushbuf(ls);
@@ -465,19 +482,22 @@ FAST_CALL void listing_line_cut2(Listing *ls, linecpos_t pos) {
     }
 }
 
-FAST_CALL void listing_set_cpumode(Listing *ls, const struct cpu_s *cpumode) {
+FAST_CALL void listing_set_cpumode(const struct cpu_s *cpumode) {
+    Listing *const ls = listing;
     if (ls == NULL) return;
     ls->disasm = cpumode->disasm;
     ls->mnemonic = cpumode->mnemonic;
 }
 
-void listing_instr(Listing *ls, unsigned int cod, uint32_t adr, int ln) {
+void listing_instr(unsigned int cod, uint32_t adr, int ln) {
+    Listing *const ls = listing;
     address_t addr, addr2;
-    if (nolisting != 0 || temporary_label_branch != 0) return;
+    if (nolisting != 0 || in_function) return;
     if (ls == NULL) {
         if (!fixeddig || constcreated || listing_pccolumn) return;
-        addr = ((current_address->l_address.address - ln - 1) & 0xffff) | current_address->l_address.bank;
-        addr2 = (current_address->address - ln - 1) & all_mem2;
+        ln++;
+        addr = (current_address->l_address - (unsigned int)ln) & all_mem;
+        addr2 = (current_address->address - (unsigned int)ln) & all_mem2;
         if (addr2 != addr) listing_pccolumn = true;
         return;
     }
@@ -486,9 +506,10 @@ void listing_instr(Listing *ls, unsigned int cod, uint32_t adr, int ln) {
         if (llist != NULL) printline(ls);
         padding2(ls, ls->columns.addr);
     }
-    addr = ((current_address->l_address.address - ln - 1) & 0xffff) | current_address->l_address.bank;
-    addr2 = (current_address->address - ln - 1) & all_mem2;
-    printaddr(ls, '.', addr2, addr);
+    addr = (current_address->l_address - (unsigned int)(ln + 1)) & all_mem;
+    addr2 = (current_address->address - (unsigned int)(ln + 1)) & all_mem2;
+    *ls->s++ = '.';
+    printaddr2(ls, addr2, addr);
     if (ln >= 0) {
         printhex(ls, cod ^ outputeor, adr ^ outputeor, ln);
         if (ls->monitor) {
@@ -499,7 +520,8 @@ void listing_instr(Listing *ls, unsigned int cod, uint32_t adr, int ln) {
     newline(ls);
 }
 
-void listing_mem(Listing *ls, const uint8_t *data, size_t len, address_t myaddr, address_t myaddr2) {
+void listing_mem(const uint8_t *data, size_t len, address_t myaddr, address_t myaddr2) {
+    Listing *const ls = listing;
     bool print, exitnow;
     int lcol;
     unsigned int repeat;
@@ -510,7 +532,7 @@ void listing_mem(Listing *ls, const uint8_t *data, size_t len, address_t myaddr,
     } prev, current;
     size_t p;
 
-    if (nolisting != 0 || temporary_label_branch != 0) return;
+    if (nolisting != 0 || in_function) return;
     if (ls == NULL) {
          if (myaddr != myaddr2) listing_pccolumn = true;
          return;
@@ -528,11 +550,12 @@ void listing_mem(Listing *ls, const uint8_t *data, size_t len, address_t myaddr,
                     if (repeat != 0) {
                         if (ls->linenum) padding2(ls, ls->columns.addr);
                         if (repeat == 1) {
-                            printaddr(ls, '>', prev.addr, prev.addr2);
+                            *ls->s++ = '>';
+                            printaddr2(ls, prev.addr, prev.addr2);
                             printhex2(ls, prev.len, prev.data);
                             flushbuf(ls);
                         } else {
-                            ls->s[ls->i++] = ';';
+                            *ls->s++ = ';';
                             padding2(ls, ls->columns.hex);
                             flushbuf(ls);
                             fprintf(ls->flist, "...repeated %u times (%u bytes)...", repeat, repeat * 16);
@@ -545,7 +568,8 @@ void listing_mem(Listing *ls, const uint8_t *data, size_t len, address_t myaddr,
                             if (print) printline(ls);
                             padding2(ls, ls->columns.addr);
                         }
-                        printaddr(ls, '>', current.addr, current.addr2);
+                        *ls->s++ = '>';
+                        printaddr2(ls, current.addr, current.addr2);
                         if (current.len != 0) {
                             printhex2(ls, current.len, current.data);
                         }
@@ -567,7 +591,7 @@ void listing_mem(Listing *ls, const uint8_t *data, size_t len, address_t myaddr,
             }
             current.data[current.len++] = data[p++];
             myaddr = (myaddr + 1) & all_mem2;
-            myaddr2 = ((myaddr2 + 1) & 0xffff) | (myaddr2 & ~(address_t)0xffff);
+            myaddr2 = (myaddr2 + 1) & all_mem;
             len--;
         }
     }
@@ -575,7 +599,8 @@ void listing_mem(Listing *ls, const uint8_t *data, size_t len, address_t myaddr,
     goto flush;
 }
 
-void listing_file(Listing *ls, const char *txt, const struct file_s *file) {
+void listing_file(const char *txt, const struct file_s *file) {
+    Listing *const ls = listing;
     if (ls == NULL) return;
     newline(ls);
     if (ls->linenum) {
@@ -590,7 +615,7 @@ void listing_file(Listing *ls, const char *txt, const struct file_s *file) {
         flushbuf(ls);
     };
     fputs(txt, ls->flist);
-    if (file != NULL) argv_print(file->realname, ls->flist);
+    if (file != NULL) argv_print(file->name, ls->flist);
     newline(ls);
     newline(ls);
 }
