@@ -1,5 +1,5 @@
 /*
-    $Id: listobj.h 2137 2020-01-02 00:52:17Z soci $
+    $Id: listobj.h 2737 2021-10-06 20:50:52Z soci $
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -24,7 +24,6 @@
 
 extern struct Type *const LIST_OBJ;
 extern struct Type *const TUPLE_OBJ;
-extern struct Type *const ADDRLIST_OBJ;
 extern struct Type *const COLONLIST_OBJ;
 
 typedef struct List {
@@ -40,43 +39,40 @@ typedef struct List {
     } u;
 } List;
 typedef struct List Tuple;
-typedef struct List Addrlist;
 typedef struct List Colonlist;
 
-extern Tuple *null_tuple;
-extern List *null_list;
-extern Addrlist *null_addrlist;
+#define List(a) OBJ_CAST(List, a)
+#define Tuple(a) OBJ_CAST(Tuple, a)
+#define Colonlist(a) OBJ_CAST(Colonlist, a)
+
+extern Obj *const null_tuple;
+extern Obj *const null_list;
 
 extern void listobj_init(void);
 extern void listobj_names(void);
 extern void listobj_destroy(void);
 
-static inline List *ref_list(List *v1) {
-    v1->v.refcount++; return v1;
-}
-
-static inline Tuple *ref_tuple(Tuple *v1) {
-    v1->v.refcount++; return v1;
-}
-
-static inline Addrlist *ref_addrlist(Addrlist *v1) {
-    v1->v.refcount++; return v1;
-}
-
-static inline Colonlist *ref_colonlist(Colonlist *v1) {
-    v1->v.refcount++; return v1;
-}
-
 static inline MUST_CHECK List *new_list(void) {
-    return (List *)val_alloc(LIST_OBJ);
-}
-static inline MUST_CHECK Addrlist *new_addrlist(void) {
-    return (Addrlist *)val_alloc(ADDRLIST_OBJ);
+    return List(val_alloc(LIST_OBJ));
 }
 static inline MUST_CHECK Colonlist *new_colonlist(void) {
-    return (Colonlist *)val_alloc(COLONLIST_OBJ);
+    return Colonlist(val_alloc(COLONLIST_OBJ));
 }
 
+struct sliceparam_s {
+    uval_t length;
+    ival_t offset, end, step;
+};
+
+struct indexoffs_s {
+    Obj *val;
+    size_t len;
+    size_t offs; 
+    linepos_t epoint;
+};
+
+extern MUST_CHECK Obj *indexoffs(struct indexoffs_s *);
+extern MUST_CHECK Obj *sliceparams(struct sliceparam_s *, const struct indexoffs_s *);
 extern MUST_CHECK Tuple *new_tuple(size_t);
 extern Obj **list_create_elements(List *, size_t);
 extern MUST_CHECK bool list_extend(List *);
