@@ -1,5 +1,5 @@
 /*
-    $Id: mem.c 2197 2020-04-06 19:00:55Z soci $
+    $Id: mem.c 2967 2023-01-08 22:50:27Z soci $
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -18,36 +18,33 @@
 */
 #include "mem.h"
 #include <string.h>
-#ifdef _WIN32
-#include <fcntl.h>
-#endif
-#ifdef __DJGPP__
+#if defined _WIN32 || defined __MSDOS__ || defined __DOS__
 #include <io.h>
 #include <fcntl.h>
+#ifndef STDOUT_FILENO
+#define STDOUT_FILENO 1
+#endif
 #endif
 #include <errno.h>
 #include "error.h"
-#include "file.h"
+#include "unicode.h"
 #include "64tass.h"
 #include "listing.h"
 #include "arguments.h"
 #include "values.h"
-
 #include "memblocksobj.h"
 
 static int memblockcomp(const void *a, const void *b) {
-    const struct memblock_s *aa = (const struct memblock_s *)a;
-    const struct memblock_s *bb = (const struct memblock_s *)b;
-    address_t ad = aa->addr, bd = bb->addr;
-    if (ad < bd) return -1;
-    return (ad > bd) ? 1 : 0;
+    address_t aa = ((const struct memblock_s *)a)->addr;
+    address_t bb = ((const struct memblock_s *)b)->addr;
+    if (aa < bb) return -1;
+    return (aa > bb) ? 1 : 0;
 }
 
 static void memcomp(Memblocks *memblocks, bool nomerge) {
-    unsigned int i, j, k;
+    size_t i, j, k;
     if (!memblocks->flattened) {
         memblocks->flattened = true;
-        memjmp(memblocks, 0);
 
         for (j = 0; j < memblocks->p; j++) { /* replace references with real copies */
             Memblocks *b = memblocks->data[j].ref;
@@ -57,9 +54,8 @@ static void memcomp(Memblocks *memblocks, bool nomerge) {
                 rest = memblocks->p - j - 1;
                 memblocks->p += b->p - 1;
                 if (memblocks->p >= memblocks->len) {
-                    memblocks->len = memblocks->p + 64;
-                    if (/*memblocks->len < 64 ||*/ memblocks->len > SIZE_MAX / sizeof *memblocks->data) err_msg_out_of_memory(); /* overflow */
-                    memblocks->data = (struct memblock_s *)reallocx(memblocks->data, memblocks->len * sizeof *memblocks->data);
+                    if (add_overflow(memblocks->p, 64, &memblocks->len)) err_msg_out_of_memory();
+                    resize_array(&memblocks->data, memblocks->len);
                 }
                 memmove(&memblocks->data[j + b->p], &memblocks->data[j + 1], rest * sizeof *memblocks->data);
                 for (k = 0; k < b->p; k++) {
@@ -70,14 +66,13 @@ static void memcomp(Memblocks *memblocks, bool nomerge) {
                     b2->addr = b->data[k].addr;
                     memblocks->mem.p += b2->len;
                     if (memblocks->mem.p >= memblocks->mem.len) {
-                        memblocks->mem.len = memblocks->mem.p + 0x1000;
-                        if (memblocks->mem.len < 0x1000) err_msg_out_of_memory(); /* overflow */
-                        memblocks->mem.data = (uint8_t *)reallocx(memblocks->mem.data, memblocks->mem.len);
+                        if (add_overflow(memblocks->mem.p, 0x1000, &memblocks->mem.len)) err_msg_out_of_memory();
+                        resize_array(&memblocks->mem.data, memblocks->mem.len);
                     }
                     memcpy(&memblocks->mem.data[b2->p], &b->mem.data[b->data[k].p], b2->len);
                 }
                 j--;
-                val_destroy(&b->v);
+                val_destroy(Obj(b));
             }
         }
     }
@@ -90,7 +85,7 @@ static void memcomp(Memblocks *memblocks, bool nomerge) {
             for (i = j + 1; i < memblocks->p; i++) if (memblocks->data[i].len != 0) {
                 struct memblock_s *bi = &memblocks->data[i];
                 if (bj->addr <= bi->addr && (bj->addr + bj->len) > bi->addr) {
-                    size_t overlap = (bj->addr + bj->len) - bi->addr;
+                    address_t overlap = (bj->addr + bj->len) - bi->addr;
                     if (overlap > bi->len) overlap = bi->len;
                     memcpy(memblocks->mem.data + bj->p + (bi->addr - bj->addr), memblocks->mem.data + bi->p, overlap);
                     bi->len -= overlap;
@@ -99,7 +94,7 @@ static void memcomp(Memblocks *memblocks, bool nomerge) {
                     continue;
                 }
                 if (bi->addr <= bj->addr && (bi->addr + bi->len) > bj->addr) {
-                    size_t overlap = bi->addr + bi->len - bj->addr;
+                    address_t overlap = bi->addr + bi->len - bj->addr;
                     if (overlap > bj->len) overlap = bj->len;
                     bj->addr += overlap;
                     bj->p += overlap;
@@ -123,13 +118,9 @@ void memjmp(Memblocks *memblocks, address_t adr) {
         memblocks->lastaddr = adr;
         return;
     }
-    if (memblocks->p >= memblocks->len) {
-        memblocks->len += 64;
-        if (/*memblocks->len < 64 ||*/ memblocks->len > SIZE_MAX / sizeof *memblocks->data) err_msg_out_of_memory(); /* overflow */
-        memblocks->data = (struct memblock_s *)reallocx(memblocks->data, memblocks->len * sizeof *memblocks->data);
-    }
+    if (memblocks->p >= memblocks->len) extend_array(&memblocks->data, &memblocks->len, 64);
     block = &memblocks->data[memblocks->p++];
-    block->len = (address_t)(memblocks->mem.p - memblocks->lastp);
+    block->len = memblocks->mem.p - memblocks->lastp;
     block->p = memblocks->lastp;
     block->ref = NULL;
     block->addr = memblocks->lastaddr;
@@ -137,84 +128,77 @@ void memjmp(Memblocks *memblocks, address_t adr) {
     memblocks->lastaddr = adr;
 }
 
-void memref(Memblocks *memblocks, Memblocks *ref) {
+void memclose(Memblocks *memblocks) {
+    size_t j;
+    for (j = 0; j < memblocks->p; j++) {
+        Memblocks *b = memblocks->data[j].ref;
+        if (b != NULL) memclose(b);
+    }
+    memjmp(memblocks, 0);
+}
+
+void memref(Memblocks *memblocks, Memblocks *ref, address_t addr, address_t ln) {
     struct memblock_s *block;
-    if (memblocks->p >= memblocks->len) {
-        memblocks->len += 64;
-        if (/*memblocks->len < 64 ||*/ memblocks->len > SIZE_MAX / sizeof *memblocks->data) err_msg_out_of_memory(); /* overflow */
-        memblocks->data = (struct memblock_s *)reallocx(memblocks->data, memblocks->len * sizeof *memblocks->data);
-    }
+    if (memblocks->p >= memblocks->len) extend_array(&memblocks->data, &memblocks->len, 64);
     block = &memblocks->data[memblocks->p++];
-    block->len = 0;
+    block->len = ln;
     block->p = memblocks->lastp;
-    block->ref = (Memblocks *)val_reference(&ref->v);
-    block->addr = memblocks->lastaddr;
+    block->ref = Memblocks(val_reference(Obj(ref)));
+    block->addr = addr;
 }
 
-void memprint(Memblocks *memblocks) {
-    size_t i;
-    char temp[10], temp2[10];
-
-    memcomp(memblocks, false);
-
-    for (i = 0; i < memblocks->p;) {
-        const struct memblock_s *block = &memblocks->data[i];
-        address_t start = block->addr;
-        address_t size = block->len;
-        for (i++; i < memblocks->p; i++) {
-            const struct memblock_s *b = &memblocks->data[i];
-            address_t addr = start + size;
-            if (b->addr != addr || addr < start) break;
-            size += b->len;
+#if defined _POSIX_C_SOURCE || defined __unix__
+static MUST_CHECK bool padding(FILE *f, address_t size, bool append)
+#else
+static MUST_CHECK bool padding(FILE *f, address_t size, bool UNUSED(append))
+#endif
+{
+    uint8_t nuls[256];
+#if defined _POSIX_C_SOURCE || defined __unix__
+    if (!append) {
+        while (size >= 0x80000000) {
+            if (fseek(f, 0x40000000, SEEK_CUR) != 0) goto err;
+            size -= 0x40000000;
+        }
+        if ((long)size > 256 && fseek(f, (long)size, SEEK_CUR) == 0) {
+            return false;
         }
-        sprintf(temp, "$%04" PRIaddress, start);
-        sprintf(temp2, "$%04" PRIaddress, (address_t)(start + size - 1));
-        printf("Memory range: %10s-%-7s $%04" PRIaddress "\n", temp, temp2, size);
-    }
-}
-
-static void padding(size_t size, FILE *f) {
-    unsigned char nuls[256];
-    while (size >= 0x80000000) {
-        if (fseek(f, 0x40000000, SEEK_CUR) != 0) goto err;
-        size -= 0x40000000;
-    }
-    if ((long)size > 256 && fseek(f, (long)size, SEEK_CUR) == 0) {
-        return;
     }
 err:
+#endif
     nuls[0] = 1;
     while (size != 0) {
-        size_t db = size < sizeof nuls ? size : sizeof nuls;
+        address_t db = size < sizeof nuls ? size : sizeof nuls;
         size -= db;
         if (nuls[0] != 0) memset(nuls, 0, db);
-        if (fwrite(nuls, db, 1, f) == 0) return;
+        if (fwrite(nuls, db, 1, f) == 0) return true;
     }
+    return false;
 }
 
 static void output_mem_c64(FILE *fout, const Memblocks *memblocks, const struct output_s *output) {
     address_t pos, end;
-    unsigned int i;
+    size_t i;
     unsigned char header[4];
 
     if (memblocks->p == 0) return;
     pos = memblocks->data[0].addr;
     switch (output->mode) {
     case OUTPUT_CBM:
-        header[0] = pos;
-        header[1] = pos >> 8;
+        header[0] = (uint8_t)pos;
+        header[1] = (uint8_t)(pos >> 8);
         if (output->longaddr) {
-            header[2] = pos >> 16;
+            header[2] = (uint8_t)(pos >> 16);
             i = 3;
         } else i = 2;
         break;
     case OUTPUT_APPLE:
-        header[0] = pos;
-        header[1] = pos >> 8;
+        header[0] = (uint8_t)pos;
+        header[1] = (uint8_t)(pos >> 8);
         end = memblocks->data[memblocks->p - 1].addr + memblocks->data[memblocks->p - 1].len;
         end -= pos;
-        header[2] = end;
-        header[3] = end >> 8;
+        header[2] = (uint8_t)end;
+        header[3] = (uint8_t)(end >> 8);
         i = 4;
         break;
     default:
@@ -224,7 +208,7 @@ static void output_mem_c64(FILE *fout, const Memblocks *memblocks, const struct
     if (i != 0 && fwrite(header, i, 1, fout) == 0) return;
     for (i = 0; i < memblocks->p; i++) {
         const struct memblock_s *block = &memblocks->data[i];
-        padding(block->addr - pos, fout);
+        if (padding(fout, block->addr - pos, output->append)) return;
         if (fwrite(memblocks->mem.data + block->p, block->len, 1, fout) == 0) return;
         pos = block->addr + block->len;
     }
@@ -243,16 +227,16 @@ static void output_mem_nonlinear(FILE *fout, const Memblocks *memblocks, bool lo
             if (b->addr != addr || addr < start) break;
             size += b->len;
         }
-        header[0] = size;
-        header[1] = size >> 8;
+        header[0] = (uint8_t)size;
+        header[1] = (uint8_t)(size >> 8);
         if (longaddr) {
-            header[2] = size >> 16;
-            header[3] = start;
-            header[4] = start >> 8;
-            header[5] = start >> 16;
+            header[2] = (uint8_t)(size >> 16);
+            header[3] = (uint8_t)start;
+            header[4] = (uint8_t)(start >> 8);
+            header[5] = (uint8_t)(start >> 16);
         } else {
-            header[2] = start;
-            header[3] = start >> 8;
+            header[2] = (uint8_t)start;
+            header[3] = (uint8_t)(start >> 8);
         }
         if (fwrite(header, longaddr ? 6 : 4, 1, fout) == 0) return;
         for (;i < j; i++) {
@@ -264,13 +248,13 @@ static void output_mem_nonlinear(FILE *fout, const Memblocks *memblocks, bool lo
     fwrite(header, longaddr ? 3 : 2, 1, fout);
 }
 
-static void output_mem_flat(FILE *fout, const Memblocks *memblocks) {
-    address_t pos;
+static void output_mem_flat(FILE *fout, const Memblocks *memblocks, bool append) {
+    address_t pos = 0;
     size_t i;
 
-    for (pos = i = 0; i < memblocks->p; i++) {
+    for (i = 0; i < memblocks->p; i++) {
         const struct memblock_s *block = &memblocks->data[i];
-        padding(block->addr - pos, fout);
+        if (padding(fout, block->addr - pos, append)) return;
         if (fwrite(memblocks->mem.data + block->p, block->len, 1, fout) == 0) return;
         pos = block->addr + block->len;
     }
@@ -293,11 +277,11 @@ static void output_mem_atari_xex(FILE *fout, const Memblocks *memblocks) {
             size += b->len;
         }
         special = (i == 0 || start == 0xffff);
-        header[2] = start;
-        header[3] = start >> 8;
+        header[2] = (uint8_t)start;
+        header[3] = (uint8_t)(start >> 8);
         end = start + size - 1;
-        header[4] = end;
-        header[5] = end >> 8;
+        header[4] = (uint8_t)end;
+        header[5] = (uint8_t)(end >> 8);
         if (fwrite(special ? header : &header[2], special ? 6 : 4, 1, fout) == 0) return;
         for (;i < j; i++) {
             const struct memblock_s *b = &memblocks->data[i];
@@ -313,42 +297,46 @@ struct hexput_s {
 
 static void hexput(struct hexput_s *h, unsigned int b) {
     const char *hex = "0123456789ABCDEF";
-    *h->line++ = hex[b >> 4];
-    *h->line++ = hex[b & 0xf];
     h->sum += b;
+    *h->line++ = hex[(b >> 4) & 0xf];
+    *h->line++ = hex[b & 0xf];
 }
 
+enum { IHEX_LENGTH = 32U };
+
 struct ihex_s {
     FILE *file;
     address_t address, segment;
-    uint8_t data[32];
+    uint8_t data[IHEX_LENGTH];
     unsigned int length;
 };
 
-static void output_mem_ihex_line(struct ihex_s *ihex, unsigned int length, address_t address, unsigned int type, const uint8_t *data) {
+static MUST_CHECK bool output_mem_ihex_line(struct ihex_s *ihex, unsigned int length, address_t address, unsigned int type, const uint8_t *data) {
     unsigned int i;
-    char line[1+(1+2+1+32+1)*2+1];
-    struct hexput_s h = { line + 1, 0 };
+    char line[1+(1+2+1+IHEX_LENGTH+1)*2+1];
+    struct hexput_s h;
+    h.line = line + 1;
+    h.sum =  0;
     line[0] = ':';
     hexput(&h, length);
-    hexput(&h, (address >> 8) & 0xff);
-    hexput(&h, address & 0xff);
+    hexput(&h, address >> 8);
+    hexput(&h, address);
     hexput(&h, type);
     for (i = 0; i < length; i++) {
         hexput(&h, data[i]);
     }
-    hexput(&h, (-h.sum) & 0xff);
+    hexput(&h, -h.sum);
     *h.line++ = '\n';
-    fwrite(line, h.line - line, 1, ihex->file);
+    return fwrite(line, (size_t)(h.line - line), 1, ihex->file) == 0;
 }
 
-static void output_mem_ihex_data(struct ihex_s *ihex) {
+static MUST_CHECK bool output_mem_ihex_data(struct ihex_s *ihex) {
     const uint8_t *data = ihex->data;
     if ((ihex->address & 0xffff) + ihex->length > 0x10000) {
         uint16_t length = (uint16_t)-ihex->address;
         unsigned int remains = ihex->length - length;
         ihex->length = length;
-        output_mem_ihex_data(ihex);
+        if (output_mem_ihex_data(ihex)) return true;
         data += length;
         ihex->length = remains;
     }
@@ -356,12 +344,13 @@ static void output_mem_ihex_data(struct ihex_s *ihex) {
         uint8_t ez[2];
         ez[0] = (uint8_t)(ihex->address >> 24);
         ez[1] = (uint8_t)(ihex->address >> 16);
-        output_mem_ihex_line(ihex, sizeof ez, 0, 4, ez);
+        if (output_mem_ihex_line(ihex, sizeof ez, 0, 4, ez)) return true;
         ihex->segment = ihex->address;
     }
-    output_mem_ihex_line(ihex, ihex->length, ihex->address, 0, data);
+    if (output_mem_ihex_line(ihex, ihex->length, ihex->address, 0, data)) return true;
     ihex->address += ihex->length;
     ihex->length = 0;
+    return false;
 }
 
 static void output_mem_ihex(FILE *fout, const Memblocks *memblocks) {
@@ -374,72 +363,9 @@ static void output_mem_ihex(FILE *fout, const Memblocks *memblocks) {
     ihex.length = 0;
     for (i = 0; i < memblocks->p; i++) {
         const struct memblock_s *b = &memblocks->data[i];
-        const uint8_t *d = memblocks->mem.data + b->p;
-        address_t addr = b->addr;
-        size_t blen = b->len;
-        if (blen != 0 && ihex.address + ihex.length != addr) {
-            if (ihex.length != 0) output_mem_ihex_data(&ihex);
-            ihex.address = addr;
-        }
-        while (blen != 0) {
-            size_t left = (sizeof ihex.data) - ihex.length;
-            size_t copy = blen > left ? left : blen;
-            memcpy(ihex.data + ihex.length, d, copy);
-            ihex.length += copy;
-            d += copy;
-            blen -= copy;
-            if (ihex.length == sizeof ihex.data) {
-                output_mem_ihex_data(&ihex);
-            }
-        }
-    }
-    if (ihex.length != 0) output_mem_ihex_data(&ihex);
-    output_mem_ihex_line(&ihex, 0, 0, 1, NULL);
-}
-
-struct srecord_s {
-    FILE *file;
-    unsigned int type;
-    address_t address;
-    uint8_t data[32];
-    unsigned int length;
-};
-
-static void output_mem_srec_line(struct srecord_s *srec) {
-    unsigned int i;
-    char line[1+1+(1+4+32+1)*2+1];
-    struct hexput_s h = { line + 2, 0 };
-    line[0] = 'S';
-    line[1] = srec->length != 0 ? ('1' + srec->type) : ('9' - srec->type);
-    hexput(&h, srec->length + srec->type + 3);
-    if (srec->type > 1) hexput(&h, (srec->address >> 24) & 0xff);
-    if (srec->type > 0) hexput(&h, (srec->address >> 16) & 0xff);
-    hexput(&h, (srec->address >> 8) & 0xff);
-    hexput(&h, srec->address & 0xff);
-    for (i = 0; i < srec->length; i++) {
-        hexput(&h, srec->data[i]);
-    }
-    hexput(&h, (~h.sum) & 0xff);
-    *h.line++ = '\n';
-    srec->address += srec->length;
-    srec->length = 0;
-    fwrite(line, h.line - line, 1, srec->file);
-}
-
-static void output_mem_srec(FILE *fout, const Memblocks *memblocks) {
-    struct srecord_s srec;
-    size_t i;
-
-    srec.file = fout;
-    srec.type = 0;
-    srec.address = 0;
-    srec.length = 0;
-    for (i = 0; i < memblocks->p; i++) {
-        const struct memblock_s *b = &memblocks->data[i];
-        size_t end = b->addr + b->len - 1;
-        if (end >= 0x10000 && srec.type < 1) srec.type = 1;
-        if (end >= 0x1000000 && srec.type < 2) {
-            srec.type = 2;
+        address_t end = b->addr + b->len - 1;
+        if (end >= 0x10000) {
+            ihex.segment = ~memblocks->data[0].addr;
             break;
         }
     }
@@ -447,56 +373,248 @@ static void output_mem_srec(FILE *fout, const Memblocks *memblocks) {
         const struct memblock_s *b = &memblocks->data[i];
         const uint8_t *d = memblocks->mem.data + b->p;
         address_t addr = b->addr;
-        size_t blen = b->len;
+        address_t blen = b->len;
+        if (blen != 0 && ihex.address + ihex.length != addr) {
+            if (ihex.length != 0) {
+                if (output_mem_ihex_data(&ihex)) return;
+            }
+            ihex.address = addr;
+        }
+        while (blen != 0) {
+            unsigned int left = IHEX_LENGTH - ihex.length;
+            address_t copy = blen > left ? left : blen;
+            memcpy(ihex.data + ihex.length, d, copy);
+            ihex.length += copy;
+            d += copy;
+            blen -= copy;
+            if (ihex.length == sizeof ihex.data) {
+                if (output_mem_ihex_data(&ihex)) return;
+            }
+        }
+    }
+    if (ihex.length != 0) {
+        if (output_mem_ihex_data(&ihex)) return;
+    }
+    if (output_mem_ihex_line(&ihex, 0, 0, 1, NULL)) return;
+}
+
+enum { MHEX_LENGTH = 24U };
+
+struct mhex_s {
+    FILE *file;
+    address_t address;
+    unsigned int lines;
+    uint8_t data[MHEX_LENGTH];
+    unsigned int length;
+};
+
+static MUST_CHECK bool output_mem_mhex_line(struct mhex_s *mhex, unsigned int length, address_t address, const uint8_t *data) {
+    unsigned int i;
+    char line[1+(1+2+MHEX_LENGTH+2)*2+2];
+    struct hexput_s h;
+    h.line = line + 1;
+    h.sum =  0;
+    line[0] = ';';
+    hexput(&h, length);
+    hexput(&h, address >> 8);
+    hexput(&h, address & 0xff);
+    for (i = 0; i < length; i++) {
+        hexput(&h, data[i]);
+    }
+    i = h.sum;
+    hexput(&h, i >> 8);
+    hexput(&h, i);
+    *h.line++ = '\r';
+    *h.line++ = '\n';
+    return fwrite(line, (size_t)(h.line - line), 1, mhex->file) == 0;
+}
+
+static MUST_CHECK bool output_mem_mhex_data(struct mhex_s *mhex) {
+    const uint8_t *data = mhex->data;
+    if ((mhex->address & 0xffff) + mhex->length > 0x10000) {
+        uint16_t length = (uint16_t)-mhex->address;
+        unsigned int remains = mhex->length - length;
+        mhex->length = length;
+        if (output_mem_mhex_data(mhex)) return true;
+        data += length;
+        mhex->length = remains;
+    }
+    if (output_mem_mhex_line(mhex, mhex->length, mhex->address, data)) return true;
+    mhex->lines++;
+    mhex->address += mhex->length;
+    mhex->length = 0;
+    return false;
+}
+
+static void output_mem_mhex(FILE *fout, const Memblocks *memblocks) {
+    struct mhex_s mhex;
+    size_t i;
+
+    mhex.file = fout;
+    mhex.address = 0;
+    mhex.lines = 0;
+    mhex.length = 0;
+    for (i = 0; i < memblocks->p; i++) {
+        const struct memblock_s *b = &memblocks->data[i];
+        const uint8_t *d = memblocks->mem.data + b->p;
+        address_t addr = b->addr;
+        address_t blen = b->len;
+        if (blen != 0 && mhex.address + mhex.length != addr) {
+            if (mhex.length != 0) {
+                if (output_mem_mhex_data(&mhex)) return;
+            }
+            mhex.address = addr;
+        }
+        while (blen != 0) {
+            unsigned int left = MHEX_LENGTH - mhex.length;
+            address_t copy = blen > left ? left : blen;
+            memcpy(mhex.data + mhex.length, d, copy);
+            mhex.length += copy;
+            d += copy;
+            blen -= copy;
+            if (mhex.length == sizeof mhex.data) {
+                if (output_mem_mhex_data(&mhex)) return;
+            }
+        }
+    }
+    if (mhex.length != 0) {
+        if (output_mem_mhex_data(&mhex)) return;
+    }
+    if (output_mem_mhex_line(&mhex, 0, mhex.lines, NULL)) return;
+}
+
+enum { SRECORD_LENGTH = 32U };
+
+struct srecord_s {
+    FILE *file;
+    unsigned int count;
+    char rectype;
+    unsigned int addrtype;
+    unsigned int length;
+    address_t address;
+    uint8_t data[SRECORD_LENGTH];
+};
+
+static MUST_CHECK bool output_mem_srec_line(struct srecord_s *srec) {
+    unsigned int i;
+    char line[1+1+(1+4+SRECORD_LENGTH+1)*2+1];
+    struct hexput_s h;
+    h.line = line + 2;
+    h.sum = 0;
+    line[0] = 'S';
+    line[1] = srec->rectype;
+    hexput(&h, srec->length + srec->addrtype + 3);
+    if (srec->addrtype > 1) hexput(&h, srec->address >> 24);
+    if (srec->addrtype > 0) hexput(&h, srec->address >> 16);
+    hexput(&h, srec->address >> 8);
+    hexput(&h, srec->address);
+    for (i = 0; i < srec->length; i++) {
+        hexput(&h, srec->data[i]);
+    }
+    hexput(&h, ~h.sum);
+    *h.line++ = '\n';
+    srec->address += srec->length;
+    srec->length = 0;
+    srec->count++;
+    return fwrite(line, (size_t)(h.line - line), 1, srec->file) == 0;
+}
+
+static void output_mem_srec(FILE *fout, const Memblocks *memblocks) {
+    struct srecord_s srec;
+    size_t i;
+    unsigned int addrtype;
+
+    srec.file = fout;
+    srec.count = 0;
+    srec.addrtype = 0;
+    srec.address = 0;
+    srec.length = 3;
+    srec.rectype = '0';
+    srec.data[0] = 'H';
+    srec.data[1] = 'D';
+    srec.data[2] = 'R';
+    if (output_mem_srec_line(&srec)) return;
+    addrtype = 0;
+    for (i = 0; i < memblocks->p; i++) {
+        const struct memblock_s *b = &memblocks->data[i];
+        address_t end = b->addr + b->len - 1;
+        if (end >= 0x10000 && addrtype < 1) addrtype = 1;
+        if (end >= 0x1000000 && addrtype < 2) {
+            addrtype = 2;
+            break;
+        }
+    }
+    srec.addrtype = addrtype;
+    srec.rectype = (char)('1' + addrtype);
+    for (i = 0; i < memblocks->p; i++) {
+        const struct memblock_s *b = &memblocks->data[i];
+        const uint8_t *d = memblocks->mem.data + b->p;
+        address_t addr = b->addr;
+        address_t blen = b->len;
         if (blen != 0 && srec.address + srec.length != addr) {
-            if (srec.length != 0) output_mem_srec_line(&srec);
+            if (srec.length != 0) {
+                if (output_mem_srec_line(&srec)) return;
+            }
             srec.address = addr;
         }
         while (blen != 0) {
-            size_t left = (sizeof srec.data) - srec.length;
-            size_t copy = blen > left ? left : blen;
+            unsigned int left = SRECORD_LENGTH - srec.length;
+            address_t copy = blen > left ? left : blen;
             memcpy(srec.data + srec.length, d, copy);
             srec.length += copy;
             d += copy;
             blen -= copy;
             if (srec.length == sizeof srec.data) {
-                output_mem_srec_line(&srec);
+                if (output_mem_srec_line(&srec)) return;
             }
         }
     }
-    if (srec.length != 0) output_mem_srec_line(&srec);
-    srec.address = memblocks->data[0].addr;
-    output_mem_srec_line(&srec);
+    if (srec.length != 0) {
+        if (output_mem_srec_line(&srec)) return;
+    }
+    if (srec.count <= 0x1000000) {
+        if (srec.count <= 0x10000) {
+            srec.addrtype = 0;
+            srec.rectype = '5';
+        } else {
+            srec.addrtype = 1;
+            srec.rectype = '6';
+        }
+        srec.address = srec.count - 1;
+        if (output_mem_srec_line(&srec)) return;
+    }
+    srec.addrtype = addrtype;
+    srec.rectype = (char)('9' - addrtype);
+    srec.address = (memblocks->p == 0) ? 0 : memblocks->data[0].addr;
+    if (output_mem_srec_line(&srec)) return;
 }
 
 void output_mem(Memblocks *memblocks, const struct output_s *output) {
     FILE* fout;
-    struct linepos_s nopoint = {0, 0};
     bool binary = (output->mode != OUTPUT_IHEX) && (output->mode != OUTPUT_SREC);
     int err;
-#if defined _WIN32 || defined __DJGPP__
+#if defined _WIN32 || defined __MSDOS__ || defined __DOS__
     int oldmode = -1;
 #endif
 
-    memcomp(memblocks, output->mode == OUTPUT_XEX || output->mode == OUTPUT_IHEX || output->mode == OUTPUT_SREC);
-
-    if (memblocks->mem.p == 0) return;
-
     if (dash_name(output->name)) {
-#if defined _WIN32 || defined __DJGPP__
+        if (fflush(stdout) != 0 || binary) setvbuf(stdout, NULL, binary ? _IOFBF : _IOLBF, binary ? BUFSIZ : 1024);
+#if defined _WIN32 || defined __MSDOS__ || defined __DOS__
         if (binary) oldmode = setmode(STDOUT_FILENO, O_BINARY);
 #endif
         fout = stdout;
     } else {
-        fout = file_open(output->name, binary ? "wb" : "wt");
+        fout = fopen_utf8(output->name, output->append ? (binary ? "ab" : "at") : (binary ? "wb" : "wt"));
     }
     if (fout == NULL) {
-        err_msg_file(ERROR_CANT_WRTE_OBJ, output->name, &nopoint);
+        err_msg_file2(ERROR_CANT_WRTE_OBJ, output->name);
         return;
     }
+    memcomp(memblocks, output->mode == OUTPUT_XEX || output->mode == OUTPUT_IHEX || output->mode == OUTPUT_SREC || output->mode == OUTPUT_MHEX);
+
     clearerr(fout); errno = 0;
     switch (output->mode) {
-    case OUTPUT_FLAT: output_mem_flat(fout, memblocks); break;
+    case OUTPUT_FLAT: output_mem_flat(fout, memblocks, output->append); break;
     case OUTPUT_NONLINEAR: output_mem_nonlinear(fout, memblocks, output->longaddr); break;
     case OUTPUT_XEX: output_mem_atari_xex(fout, memblocks); break;
     case OUTPUT_RAW:
@@ -504,27 +622,28 @@ void output_mem(Memblocks *memblocks, const struct output_s *output) {
     case OUTPUT_CBM: output_mem_c64(fout, memblocks, output); break;
     case OUTPUT_IHEX: output_mem_ihex(fout, memblocks); break;
     case OUTPUT_SREC: output_mem_srec(fout, memblocks); break;
+    case OUTPUT_MHEX: output_mem_mhex(fout, memblocks); break;
     }
     err = ferror(fout);
     err |= (fout != stdout) ? fclose(fout) : fflush(fout);
-    if (err != 0 && errno != 0) err_msg_file(ERROR_CANT_WRTE_OBJ, output->name, &nopoint);
-#if defined _WIN32 || defined __DJGPP__
+    if (err != 0 && errno != 0) err_msg_file2(ERROR_CANT_WRTE_OBJ, output->name);
+#if defined _WIN32 || defined __MSDOS__ || defined __DOS__
     if (oldmode >= 0) setmode(STDOUT_FILENO, oldmode);
 #endif
+    if (fout == stdout && binary) setvbuf(stdout, NULL, _IOLBF, 1024);
 }
 
-FAST_CALL uint8_t *alloc_mem(Memblocks *memblocks, size_t len) {
-    size_t p = memblocks->mem.p + len;
+FAST_CALL uint8_t *alloc_mem(Memblocks *memblocks, address_t len) {
+    address_t p;
     uint8_t *d;
-    if (p < len) err_msg_out_of_memory(); /* overflow */
+    if (add_overflow(memblocks->mem.p, len, &p)) err_msg_out_of_memory();
     if (p > memblocks->mem.len) {
-        memblocks->mem.len = p + 0x1000;
-        if (memblocks->mem.len < 0x1000) err_msg_out_of_memory(); /* overflow */
-        memblocks->mem.data = (uint8_t *)reallocx(memblocks->mem.data, memblocks->mem.len);
+        if (add_overflow(p, 0x1000, &memblocks->mem.len)) err_msg_out_of_memory();
+        resize_array(&memblocks->mem.data, memblocks->mem.len);
     }
     d = memblocks->mem.data + memblocks->mem.p;
     do {
-        size_t left = all_mem2 - (memblocks->lastaddr + memblocks->mem.p - memblocks->lastp);
+        address_t left = all_mem2 - (memblocks->lastaddr + memblocks->mem.p - memblocks->lastp);
         if (len <= left) break;
         memblocks->mem.p += left + 1;
         len -= left + 1;
@@ -534,25 +653,21 @@ FAST_CALL uint8_t *alloc_mem(Memblocks *memblocks, size_t len) {
     return d;
 }
 
-static size_t omemp;
-static size_t ptextaddr;
-static address_t oaddr, oaddr2, olastaddr;
-
-void mark_mem(const Memblocks *memblocks, address_t adr, address_t adr2) {
-    ptextaddr = memblocks->mem.p;
-    omemp = memblocks->p;
-    olastaddr = memblocks->lastaddr;
-    oaddr = adr;
-    oaddr2 = adr2;
+void mark_mem(struct mem_mark_s *mm, const Memblocks *memblocks, address_t adr, address_t adr2) {
+    mm->ptextaddr = memblocks->mem.p;
+    mm->omemp = memblocks->p;
+    mm->olastaddr = memblocks->lastaddr;
+    mm->oaddr = adr;
+    mm->oaddr2 = adr2;
 }
 
 size_t get_mem(const Memblocks *memblocks) {
     return memblocks->p;
 }
 
-int read_mem(const Memblocks *memblocks, address_t raddr, size_t membp, size_t offs) {
+int read_mem(const Memblocks *memblocks, address_t raddr, size_t membp, address_t offs) {
     address_t addr, diff;
-    size_t len;
+    address_t len;
     if (membp < memblocks->p) {
         addr = memblocks->data[membp].addr;
     } else {
@@ -581,17 +696,17 @@ int read_mem(const Memblocks *memblocks, address_t raddr, size_t membp, size_t o
     return -1;
 }
 
-void write_mark_mem(Memblocks *memblocks, unsigned int c) {
-    memblocks->mem.data[ptextaddr] = (uint8_t)c;
+void write_mark_mem(const struct mem_mark_s *mm, Memblocks *memblocks, unsigned int c) {
+    memblocks->mem.data[mm->ptextaddr] = (uint8_t)c;
 }
 
-void list_mem(const Memblocks *memblocks) {
+void list_mem(const struct mem_mark_s *mm, const Memblocks *memblocks) {
     bool first = true;
-    size_t o = omemp;
-    address_t addr2 = oaddr;
+    size_t o = mm->omemp;
+    address_t addr2 = mm->oaddr;
 
     for (; o <= memblocks->p; o++) {
-        size_t p;
+        address_t p;
         address_t addr, len;
 
         if (o < memblocks->p) {
@@ -606,7 +721,7 @@ void list_mem(const Memblocks *memblocks) {
 
         if (first) {
             first = false;
-            if (addr != olastaddr) {
+            if (addr != mm->olastaddr) {
                 len = 0;
                 o--;
             } else {
@@ -619,6 +734,6 @@ void list_mem(const Memblocks *memblocks) {
         } else {
             if (len == 0) continue;
         }
-        listing_mem(listing, memblocks->mem.data + p, len, addr, ((oaddr2 + addr - addr2) & 0xffff) | (oaddr2 & ~(address_t)0xffff));
+        listing_mem(memblocks->mem.data + p, len, addr, (mm->oaddr2 + addr - addr2) & all_mem);
     }
 }
