@@ -1,7 +1,7 @@
 /*
  *  my_getopt.c - my re-implementation of getopt.
  *  Copyright 1997, 2000, 2001, 2002, Benjamin Sittler
- *  $Id: my_getopt.c 2194 2020-04-01 08:38:40Z soci $
+ *  $Id: my_getopt.c 2951 2023-01-05 21:21:16Z soci $
  *
  *  Permission is hereby granted, free of charge, to any person
  *  obtaining a copy of this software and associated documentation
@@ -81,7 +81,7 @@ static int my_getopt(int argc, char *argv[], const char *opts)
       goto my_getopt_ok;
     }
     if (my_opterr) {fatal_error("option '-");
-                    printable_print2((const uint8_t *)argv[my_optind] + charind, stderr, ((my_optopt & 0x80) != 0) ? utf8len((unsigned int)my_optopt) : 1);
+                    printable_print2((const uint8_t *)argv[my_optind] + charind, stderr, ((my_optopt & 0x80) != 0) ? utf8len((unsigned char)my_optopt) : 1);
                     fputs("' not recognized", stderr); fatal_error(NULL); }
     opt = '?';
     if (argv[my_optind][++charind] == '\0') {
@@ -187,7 +187,7 @@ static int my_getopt_internal(int argc, char *argv[], const char *shortopts,
   } else if ((!long_only) && (argv[my_optind][1] != '-'))
     opt = my_getopt(argc, argv, shortopts);
   else {
-    int charind, offset;
+    size_t charind, offset;
     int found = 0, ind, hits = 0;
 
     if (((my_optopt = argv[my_optind][1]) != '-') && argv[my_optind][2] == 0) {
@@ -207,15 +207,19 @@ static int my_getopt_internal(int argc, char *argv[], const char *shortopts,
         (argv[my_optind][charind] != '\0') &&
           (argv[my_optind][charind] != '=');
         charind++);
-    for (ind = 0; longopts[ind].name != NULL && hits == 0; ind++)
-      if ((strlen(longopts[ind].name) == (size_t) (charind - offset)) &&
+    for (ind = 0; longopts[ind].name != NULL && hits == 0; ind++) {
+      if ((strlen(longopts[ind].name) == charind - offset) &&
          (strncmp(longopts[ind].name,
-                  argv[my_optind] + offset, charind - offset) == 0))
-        found = ind, hits++;
-    if (hits == 0) for (ind = 0; longopts[ind].name != NULL; ind++)
+                  argv[my_optind] + offset, charind - offset) == 0)) {
+        found = ind; hits++;
+      }
+    }
+    if (hits == 0) for (ind = 0; longopts[ind].name != NULL; ind++) {
       if (strncmp(longopts[ind].name,
-                 argv[my_optind] + offset, charind - offset) == 0)
-        found = ind, hits++;
+                 argv[my_optind] + offset, charind - offset) == 0) {
+        found = ind; hits++;
+      }
+    }
     if (hits == 1) {
       opt = 0;
 
