@@ -1,5 +1,5 @@
 /*
-    $Id: namespaceobj.c 2136 2020-01-01 23:01:52Z soci $
+    $Id: namespaceobj.c 2727 2021-10-03 20:21:13Z soci $
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -26,49 +26,54 @@
 
 #include "listobj.h"
 #include "strobj.h"
-#include "operobj.h"
 #include "typeobj.h"
 #include "noneobj.h"
 #include "labelobj.h"
 #include "errorobj.h"
-#include "identobj.h"
+#include "symbolobj.h"
+#include "anonsymbolobj.h"
 #include "codeobj.h"
 #include "macroobj.h"
 #include "mfuncobj.h"
+#include "boolobj.h"
 
 static Type obj;
 
 Type *const NAMESPACE_OBJ = &obj;
 
-static MUST_CHECK Obj *create(Obj *v1, linepos_t epoint) {
+static MUST_CHECK Obj *namespace_from_obj(Obj *v1, linepos_t epoint) {
     switch (v1->obj->type) {
     case T_NONE:
     case T_ERROR:
     case T_NAMESPACE: return val_reference(v1);
     case T_CODE:
-        return val_reference(&((Code *)v1)->names->v);
+        return val_reference(Obj(Code(v1)->names));
     case T_UNION:
     case T_STRUCT:
-        return val_reference(&((Struct *)v1)->names->v);
+        return val_reference(Obj(Struct(v1)->names));
     case T_MFUNC:
-        return val_reference(&((Mfunc *)v1)->names->v);
+        return val_reference(Obj(Mfunc(v1)->names));
     default: break;
     }
-    return (Obj *)new_error_conv(v1, NAMESPACE_OBJ, epoint);
+    return new_error_conv(v1, NAMESPACE_OBJ, epoint);
+}
+
+static MUST_CHECK Obj *convert(oper_t op) {
+    return namespace_from_obj(op->v2, op->epoint2);
 }
 
 static FAST_CALL void destroy(Obj *o1) {
-    Namespace *v1 = (Namespace *)o1;
+    Namespace *v1 = Namespace(o1);
     size_t i;
     if (v1->data == NULL) return;
     for (i = 0; i <= v1->mask; i++) {
-        if (v1->data[i] != NULL) val_destroy(&v1->data[i]->v);
+        if (v1->data[i] != NULL) val_destroy(Obj(v1->data[i]));
     }
     free(v1->data);
 }
 
 static FAST_CALL void garbage(Obj *o1, int j) {
-    Namespace *v1 = (Namespace *)o1;
+    Namespace *v1 = Namespace(o1);
     size_t i;
     if (v1->data == NULL) return;
     switch (j) {
@@ -82,7 +87,7 @@ static FAST_CALL void garbage(Obj *o1, int j) {
         return;
     case 1:
         for (i = 0; i <= v1->mask; i++) {
-            Obj *v = &v1->data[i]->v;
+            Obj *v = Obj(v1->data[i]);
             if (v == NULL) continue;
             if ((v->refcount & SIZE_MSB) != 0) {
                 v->refcount -= SIZE_MSB - 1;
@@ -129,7 +134,7 @@ static bool namespace_issubset(Namespace *v1, const Namespace *v2) {
                 ret = false;
                 break;
             }
-            if (!p->v.obj->same(&p->v, &p2->v)) {
+            if (!p->v.obj->same(Obj(p), Obj(p2))) {
                 ret = false;
                 break;
             }
@@ -140,13 +145,13 @@ static bool namespace_issubset(Namespace *v1, const Namespace *v2) {
 }
 
 static FAST_CALL bool same(const Obj *o1, const Obj *o2) {
-    Namespace *v1 = (Namespace *)o1, *v2 = (Namespace *)o2;
-    if (o2->obj != NAMESPACE_OBJ) return false;
+    Namespace *v1 = Namespace(o1), *v2 = Namespace(o2);
+    if (o1->obj != o2->obj) return false;
     return namespace_issubset(v1, v2) && namespace_issubset(v2, v1);
 }
 
 static MUST_CHECK Obj *repr(Obj *o1, linepos_t epoint, size_t maxsize) {
-    const Namespace *v1 = (const Namespace *)o1;
+    const Namespace *v1 = Namespace(o1);
     size_t i = 0, j, ln = 13, chars = 13;
     Obj **vals;
     Str *str;
@@ -157,8 +162,7 @@ static MUST_CHECK Obj *repr(Obj *o1, linepos_t epoint, size_t maxsize) {
     if (v1->len != 0) {
         size_t n;
         ln = v1->len;
-        chars = ln + 12;
-        if (chars < 12) return NULL; /* overflow */
+        if (add_overflow(ln, 12, &chars)) return NULL;
         if (chars > maxsize) return NULL;
         tuple = new_tuple(ln);
         vals = tuple->data;
@@ -172,11 +176,10 @@ static MUST_CHECK Obj *repr(Obj *o1, linepos_t epoint, size_t maxsize) {
                 chars--;
                 continue;
             }
-            v = p->v.obj->repr(&p->v, epoint, maxsize - chars);
+            v = p->v.obj->repr(Obj(p), epoint, maxsize - chars);
             if (v == NULL || v->obj != STR_OBJ) goto error;
-            str = (Str *)v;
-            ln += str->len;
-            if (ln < str->len) goto error2; /* overflow */
+            str = Str(v);
+            if (inc_overflow(&ln, str->len)) goto error2;
             chars += str->chars;
             if (chars > maxsize) {
             error2:
@@ -184,7 +187,7 @@ static MUST_CHECK Obj *repr(Obj *o1, linepos_t epoint, size_t maxsize) {
                 v = NULL;
             error:
                 tuple->len = i;
-                val_destroy(&tuple->v);
+                val_destroy(Obj(tuple));
                 return v;
             }
             vals[i++] = v;
@@ -194,7 +197,7 @@ static MUST_CHECK Obj *repr(Obj *o1, linepos_t epoint, size_t maxsize) {
     }
     str = new_str2(ln);
     if (str == NULL) {
-        if (tuple != NULL) val_destroy(&tuple->v);
+        if (tuple != NULL) val_destroy(Obj(tuple));
         return NULL;
     }
     str->chars = chars;
@@ -202,7 +205,7 @@ static MUST_CHECK Obj *repr(Obj *o1, linepos_t epoint, size_t maxsize) {
     memcpy(s, "namespace({", 11);
     s += 11;
     for (j = 0; j < i; j++) {
-        Str *str2 = (Str *)vals[j];
+        Str *str2 = Str(vals[j]);
         if (j != 0) *s++ = ',';
         if (str2->len != 0) {
             memcpy(s, str2->data, str2->len);
@@ -211,8 +214,8 @@ static MUST_CHECK Obj *repr(Obj *o1, linepos_t epoint, size_t maxsize) {
     }
     *s++ = '}';
     *s = ')';
-    if (tuple != NULL) val_destroy(&tuple->v);
-    return &str->v;
+    if (tuple != NULL) val_destroy(Obj(tuple));
+    return Obj(str);
 }
 
 MUST_CHECK Obj *namespace_member(oper_t op, Namespace *v1) {
@@ -220,9 +223,9 @@ MUST_CHECK Obj *namespace_member(oper_t op, Namespace *v1) {
     Error *err;
     Label *l;
     switch (o2->obj->type) {
-    case T_IDENT:
+    case T_SYMBOL:
         {
-            Ident *v2 = (Ident *)o2;
+            Symbol *v2 = Symbol(o2);
             l = find_label2(&v2->name, v1);
             if (l != NULL) {
                 if (diagnostics.case_symbol && str_cmp(&v2->name, &l->name) != 0) err_msg_symbol_case(&v2->name, l, op->epoint2);
@@ -230,41 +233,44 @@ MUST_CHECK Obj *namespace_member(oper_t op, Namespace *v1) {
                 return val_reference(l->value);
             }
             if (!referenceit || (constcreated && pass < max_pass)) {
-                return (Obj *)ref_none();
+                return ref_none();
             }
             err = new_error(ERROR___NOT_DEFINED, op->epoint2);
             err->u.notdef.names = ref_namespace(v1);
-            err->u.notdef.ident = (Obj *)ref_ident(v2);
+            err->u.notdef.symbol = Obj(ref_symbol(v2));
             err->u.notdef.down = false;
-            return &err->v;
+            return Obj(err);
         }
-    case T_ANONIDENT:
+    case T_ANONSYMBOL:
         {
-            Anonident *v2 = (Anonident *)o2;
+            Anonsymbol *v2 = Anonsymbol(o2);
             l = find_anonlabel2(v2->count, v1);
             if (l != NULL) {
                 touch_label(l);
                 return val_reference(l->value);
             }
             if (!referenceit || (constcreated && pass < max_pass)) {
-                return (Obj *)ref_none();
+                return ref_none();
             }
             err = new_error(ERROR___NOT_DEFINED, op->epoint2);
             err->u.notdef.names = ref_namespace(v1);
-            err->u.notdef.ident = (Obj *)ref_anonident(v2);
+            err->u.notdef.symbol = Obj(ref_anonsymbol(v2));
             err->u.notdef.down = false;
-            return &err->v;
+            return Obj(err);
         }
-    case T_TUPLE:
-    case T_LIST: return o2->obj->rcalc2(op);
     case T_ERROR:
     case T_NONE: return val_reference(o2);
-    default: return obj_oper_error(op);
+    default:
+        if (o2->obj->iterable) {
+            return o2->obj->rcalc2(op);
+        }
+        break;
     }
+    return obj_oper_error(op);
 }
 
 MALLOC Namespace *new_namespace(const struct file_list_s *file_list, linepos_t epoint) {
-    Namespace *val = (Namespace *)val_alloc(NAMESPACE_OBJ);
+    Namespace *val = Namespace(val_alloc(NAMESPACE_OBJ));
     val->data = NULL;
     val->mask = 0;
     val->file_list = file_list;
@@ -276,38 +282,74 @@ MALLOC Namespace *new_namespace(const struct file_list_s *file_list, linepos_t e
 }
 
 static MUST_CHECK Obj *calc2(oper_t op) {
-    if (op->op == &o_MEMBER) {
-        return namespace_member(op, (Namespace *)op->v1);
+    if (op->op == O_MEMBER) {
+        return namespace_member(op, Namespace(op->v1));
     }
-    if (op->v2 == &none_value->v || op->v2->obj == ERROR_OBJ) return val_reference(op->v2);
+    if (op->v2 == none_value || op->v2->obj == ERROR_OBJ) return val_reference(op->v2);
     return obj_oper_error(op);
 }
 
+static MUST_CHECK Obj *contains(oper_t op) {
+    Namespace *v2 = Namespace(op->v2);
+    Obj *o1 = op->v1;
+    Label *l;
+    switch (o1->obj->type) {
+    case T_SYMBOL:
+        {
+            Symbol *v1 = Symbol(o1);
+            l = find_label2(&v1->name, v2);
+            if (l != NULL) {
+                if (diagnostics.case_symbol && str_cmp(&v1->name, &l->name) != 0) err_msg_symbol_case(&v1->name, l, op->epoint);
+                if (l->constant) {
+                    err_msg_not_variable(l, &v1->name, op->epoint);
+                    l = NULL;
+                }
+            }
+        }
+        break;
+    case T_ANONSYMBOL:
+        l = find_anonlabel2(Anonsymbol(o1)->count, v2);
+        if (l != NULL && l->constant) {
+            err_msg_not_variable(l, &l->name, op->epoint);
+            l = NULL;
+        }
+        break;
+    case T_NONE:
+    case T_ERROR:
+        return val_reference(o1);
+    default:
+        return obj_oper_error(op);
+    }
+    if (l != NULL) touch_label(l);
+    return truth_reference(op->op == O_IN ? (l != NULL) : (l == NULL));
+}
+
 void namespaceobj_init(void) {
-    new_type(&obj, T_NAMESPACE, "namespace", sizeof(Namespace));
-    obj.create = create;
-    obj.destroy = destroy;
-    obj.garbage = garbage;
-    obj.same = same;
-    obj.repr = repr;
-    obj.calc2 = calc2;
+    Type *type = new_type(&obj, T_NAMESPACE, "namespace", sizeof(Namespace));
+    type->convert = convert;
+    type->destroy = destroy;
+    type->garbage = garbage;
+    type->same = same;
+    type->repr = repr;
+    type->calc2 = calc2;
+    type->contains = contains;
 }
 
 void namespaceobj_names(void) {
-    new_builtin("namespace", val_reference(&NAMESPACE_OBJ->v));
+    new_builtin("namespace", val_reference(Obj(NAMESPACE_OBJ)));
 }
 
 Namespace *get_namespace(const Obj *o) {
     switch (o->obj->type) {
     case T_CODE:
-        return ((Code *)o)->names;
+        return Code(o)->names;
     case T_UNION:
     case T_STRUCT:
-        return ((Struct *)o)->names;
+        return Struct(o)->names;
     case T_MFUNC:
-        return ((Mfunc *)o)->names;
+        return Mfunc(o)->names;
     case T_NAMESPACE:
-        return (Namespace *)o;
+        return Namespace(o);
     default:
         return NULL;
     }
