@@ -1,5 +1,5 @@
 /*
-    $Id: obj.h 1925 2019-08-25 11:09:57Z soci $
+    $Id: obj.h 2791 2022-05-25 17:55:02Z soci $
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -29,30 +29,40 @@ typedef struct Obj {
     size_t refcount;
 } Obj;
 
-typedef struct Lbl_s {
+#define Obj(a) ((0 ? &(a)->v : (Obj *)(a)))
+#define OBJ_CAST(b, a) ((b *)(1 ? (a) : (Obj *)(b *)(a)))
+
+typedef struct Lbl {
     Obj v;
-    line_t sline;
+    linenum_t sline;
+    uint8_t pass;
     size_t waitforp;
     const struct file_list_s *file_list;
     struct Namespace *parent;
 } Lbl;
 
+#define Lbl(a) OBJ_CAST(Lbl, a)
+
 typedef struct Funcargs {
     Obj v;
-    size_t len;
+    argcount_t len;
     struct values_s *val;
 } Funcargs;
 
+#define Funcargs(a) OBJ_CAST(Funcargs, a)
+
 typedef struct Default {
     Obj v;
-    int *dummy;
 } Default;
 
+#define Default(a) OBJ_CAST(Default, a)
+
 static inline Obj *val_reference(Obj *v1) {
     v1->refcount++; return v1;
 }
 
 extern MUST_CHECK Obj *obj_oper_error(struct oper_s *);
+extern MUST_CHECK Obj *obj_oper_compare(struct oper_s *, int);
 extern void obj_init(struct Type *);
 extern void objects_init(void);
 extern void objects_destroy(void);
@@ -60,9 +70,9 @@ extern void objects_destroy(void);
 extern struct Type *const LBL_OBJ;
 extern struct Type *const DEFAULT_OBJ;
 extern struct Type *const FUNCARGS_OBJ;
-extern Default *default_value;
+extern Obj *const default_value;
 
-static inline Default *ref_default(void) {
-    default_value->v.refcount++; return default_value;
+static inline Obj *ref_default(void) {
+    default_value->refcount++; return default_value;
 }
 #endif
