@@ -1,5 +1,5 @@
 /*
-    $Id: registerobj.c 2115 2019-12-11 20:27:38Z soci $
+    $Id: registerobj.c 2896 2022-11-05 05:33:41Z soci $
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -22,12 +22,12 @@
 #include "variables.h"
 #include "values.h"
 
-#include "boolobj.h"
 #include "strobj.h"
-#include "intobj.h"
-#include "operobj.h"
 #include "typeobj.h"
 #include "errorobj.h"
+#include "addressobj.h"
+#include "intobj.h"
+#include "error.h"
 
 static Type obj;
 
@@ -38,15 +38,15 @@ static FAST_CALL NO_INLINE void register_destroy(Register *v1) {
 }
 
 static FAST_CALL void destroy(Obj *o1) {
-    Register *v1 = (Register *)o1;
-    if (v1->val != v1->data) register_destroy(v1);
+    Register *v1 = Register(o1);
+    if unlikely(v1->val != v1->data) register_destroy(v1);
 }
 
 static inline MALLOC Register *new_register(void) {
-    return (Register *)val_alloc(REGISTER_OBJ);
+    return Register(val_alloc(REGISTER_OBJ));
 }
 
-static MUST_CHECK Obj *create(Obj *o1, linepos_t epoint) {
+static MUST_CHECK Obj *register_from_obj(Obj *o1, linepos_t epoint) {
     uint8_t *s;
     switch (o1->obj->type) {
     case T_NONE:
@@ -55,34 +55,38 @@ static MUST_CHECK Obj *create(Obj *o1, linepos_t epoint) {
         return val_reference(o1);
     case T_STR:
         {
-            Str *v1 = (Str *)o1;
+            Str *v1 = Str(o1);
             Register *v = new_register();
             v->chars = v1->chars;
             v->len = v1->len;
             if (v->len != 0) {
                 if (v->len > sizeof v->val) {
-                    s = (uint8_t *)malloc(v->len);
-                    if (s == NULL) return (Obj *)new_error_mem(epoint);
+                    s = allocate_array(uint8_t, v->len);
+                    if (s == NULL) return new_error_mem(epoint);
                 } else s = v->val;
                 memcpy(s, v1->data, v->len);
             } else s = NULL;
             v->data = s;
-            return &v->v;
+            return Obj(v);
         }
     default: break;
     }
-    return (Obj *)new_error_conv(o1, REGISTER_OBJ, epoint);
+    return new_error_conv(o1, REGISTER_OBJ, epoint);
+}
+
+static MUST_CHECK Obj *convert(oper_t op) {
+    return register_from_obj(op->v2, op->epoint2);
 }
 
 static FAST_CALL bool same(const Obj *o1, const Obj *o2) {
-    const Register *v1 = (const Register *)o1, *v2 = (const Register *)o2;
-    return o2->obj == REGISTER_OBJ && v1->len == v2->len && (
+    const Register *v1 = Register(o1), *v2 = Register(o2);
+    return o1->obj == o2->obj && v1->len == v2->len && (
             v1->data == v2->data ||
             memcmp(v1->data, v2->data, v2->len) == 0);
 }
 
-static MUST_CHECK Error *hash(Obj *o1, int *hs, linepos_t UNUSED(epoint)) {
-    Register *v1 = (Register *)o1;
+static MUST_CHECK Obj *hash(Obj *o1, int *hs, linepos_t UNUSED(epoint)) {
+    Register *v1 = Register(o1);
     size_t l = v1->len;
     const uint8_t *s2 = v1->data;
     unsigned int h;
@@ -92,13 +96,13 @@ static MUST_CHECK Error *hash(Obj *o1, int *hs, linepos_t UNUSED(epoint)) {
     }
     h = (unsigned int)*s2 << 7;
     while ((l--) != 0) h = (1000003 * h) ^ *s2++;
-    h ^= v1->len;
+    h ^= (unsigned int)v1->len;
     *hs = h & ((~0U) >> 1);
     return NULL;
 }
 
 static MUST_CHECK Obj *repr(Obj *o1, linepos_t UNUSED(epoint), size_t maxsize) {
-    Register *v1 = (Register *)o1;
+    Register *v1 = Register(o1);
     size_t i2, i;
     uint8_t *s, *s2;
     uint8_t q;
@@ -106,8 +110,7 @@ static MUST_CHECK Obj *repr(Obj *o1, linepos_t UNUSED(epoint), size_t maxsize) {
     Str *v;
     i = str_quoting(v1->data, v1->len, &q);
 
-    i2 = i + 12;
-    if (i2 < 12) return NULL; /* overflow */
+    if (add_overflow(i, 12, &i2)) return NULL;
     chars = i2 - (v1->len - v1->chars);
     if (chars > maxsize) return NULL;
     v = new_str2(i2);
@@ -127,11 +130,11 @@ static MUST_CHECK Obj *repr(Obj *o1, linepos_t UNUSED(epoint), size_t maxsize) {
     }
     s[i] = q;
     s[i + 1] = ')';
-    return &v->v;
+    return Obj(v);
 }
 
 static MUST_CHECK Obj *str(Obj *o1, linepos_t UNUSED(epoint), size_t maxsize) {
-    Register *v1 = (Register *)o1;
+    Register *v1 = Register(o1);
     Str *v;
 
     if (v1->chars > maxsize) return NULL;
@@ -139,46 +142,49 @@ static MUST_CHECK Obj *str(Obj *o1, linepos_t UNUSED(epoint), size_t maxsize) {
     if (v == NULL) return NULL;
     v->chars = v1->chars;
     memcpy(v->data, v1->data, v1->len);
-    return &v->v;
+    return Obj(v);
 }
 
 static inline int icmp(oper_t op) {
-    const Register *v1 = (Register *)op->v1, *v2 = (Register *)op->v2;
+    const Register *v1 = Register(op->v1), *v2 = Register(op->v2);
     int h = memcmp(v1->data, v2->data, (v1->len < v2->len) ? v1->len : v2->len);
     if (h != 0) return h;
     if (v1->len < v2->len) return -1;
     return (v1->len > v2->len) ? 1 : 0;
 }
 
-static MUST_CHECK Obj *calc2_register(oper_t op) {
-    int val = icmp(op);
-    switch (op->op->op) {
-    case O_CMP:
-        if (val < 0) return (Obj *)ref_int(minus1_value);
-        return (Obj *)ref_int(int_value[(val > 0) ? 1 : 0]);
-    case O_EQ: return truth_reference(val == 0);
-    case O_NE: return truth_reference(val != 0);
-    case O_MIN:
-    case O_LT: return truth_reference(val < 0);
-    case O_LE: return truth_reference(val <= 0);
-    case O_MAX:
-    case O_GT: return truth_reference(val > 0);
-    case O_GE: return truth_reference(val >= 0);
-    default: break;
-    }
-    return obj_oper_error(op);
-}
-
 static MUST_CHECK Obj *calc2(oper_t op) {
     const Type *t2 = op->v2->obj;
     switch (t2->type) {
     case T_REGISTER: 
-        return calc2_register(op);
+        return obj_oper_compare(op, icmp(op));
+    case T_BOOL:
+    case T_INT:
+    case T_BITS:
+    case T_FLOAT:
+    case T_BYTES:
+    case T_STR:
+    case T_CODE:
+        if (Register(op->v1)->len == 1) {
+            Address_types am = register_to_indexing(Register(op->v1)->data[0]);
+            if (am == A_NONE) break;
+            if (op->op == O_ADD) {
+                return new_address(val_reference(op->v2), am);
+            }
+            if (op->op == O_SUB) {
+                op->v1 = int_value[0];
+                op->inplace = NULL;
+                return new_address(INT_OBJ->calc2(op), am);
+            }
+        }
+        break;
+    case T_ADDRESS:
+        return t2->rcalc2(op);
     case T_NONE:
     case T_ERROR:
         return val_reference(op->v2);
     default:
-        if (t2->iterable && op->op != &o_MEMBER && op->op != &o_X) {
+        if (t2->iterable && op->op != O_MEMBER && op->op != O_X) {
             return t2->rcalc2(op);
         }
         break;
@@ -189,37 +195,51 @@ static MUST_CHECK Obj *calc2(oper_t op) {
 static MUST_CHECK Obj *rcalc2(oper_t op) {
     const Type *t1 = op->v1->obj;
     switch (t1->type) {
-    case T_NONE:
-    case T_ERROR:
-    case T_TUPLE:
-    case T_LIST:
-        if (op->op != &o_IN) {
-            return t1->calc2(op);
+    case T_BOOL:
+    case T_INT:
+    case T_BITS:
+    case T_FLOAT:
+    case T_BYTES:
+        if (Register(op->v2)->len == 1) {
+            if (op->op == O_ADD) {
+                Address_types am = register_to_indexing(Register(op->v2)->data[0]);
+                if (am != A_NONE) {
+                    return new_address(val_reference(op->v1), am);
+                }
+            }
         }
         break;
-    default: break;
+    default:
+        if (!t1->iterable) {
+            break;
+        }
+        FALL_THROUGH; /* fall through */
+    case T_NONE:
+    case T_ERROR:
+        return t1->calc2(op);
     }
     return obj_oper_error(op);
 }
 
+static uint32_t register_names;
+
 void registerobj_init(void) {
-    new_type(&obj, T_REGISTER, "register", sizeof(Register));
-    obj.create = create;
-    obj.destroy = destroy;
-    obj.same = same;
-    obj.hash = hash;
-    obj.repr = repr;
-    obj.str = str;
-    obj.calc2 = calc2;
-    obj.rcalc2 = rcalc2;
+    Type *type = new_type(&obj, T_REGISTER, "register", sizeof(Register));
+    type->convert = convert;
+    type->destroy = destroy;
+    type->same = same;
+    type->hash = hash;
+    type->repr = repr;
+    type->str = str;
+    type->calc2 = calc2;
+    type->rcalc2 = rcalc2;
+    register_names = 0;
 }
 
 void registerobj_names(void) {
-    new_builtin("register", val_reference(&REGISTER_OBJ->v));
+    new_builtin("register", val_reference(Obj(REGISTER_OBJ)));
 }
 
-static uint32_t register_names;
-
 bool registerobj_createnames(uint32_t registers) {
     uint32_t regs = registers & ~register_names;
     uint8_t name;
@@ -235,7 +255,7 @@ bool registerobj_createnames(uint32_t registers) {
         reg->data = reg->val;
         reg->len = 1;
         reg->chars = 1;
-        new_builtin((const char *)reg->val, &reg->v);
+        new_builtin((const char *)reg->val, Obj(reg));
     }
     return true;
 }
