@@ -1,5 +1,5 @@
 /*
-    $Id: section.h 2016 2019-10-20 06:41:22Z soci $
+    $Id: section.h 2853 2022-10-25 06:30:43Z soci $
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -25,16 +25,18 @@
 struct Obj;
 struct Memblocks;
 struct optimizer_s;
+struct output_s;
 
 struct section_address_s {
     address_t address;
-    address2_t l_address;
+    address_t l_address;
     address_t start;
     address_t end;
-    address2_t l_start;
-    address2_t l_union;
+    address_t l_start;
+    address_t l_union;
     bool moved;
     bool wrapwarn;
+    bool bankwarn;
     bool unionmode;
     struct Obj *l_address_val;
     struct Memblocks *mem;
@@ -50,16 +52,14 @@ struct section_s {
     uval_t conflicts;
     uval_t provides;
     address_t restart;
-    address2_t l_restart;
+    address_t l_restart;
     address_t size;
     struct section_address_s address;
     uint8_t usepass;
     uint8_t defpass;
-    uint8_t structrecursion;
     uint8_t logicalrecursion;
     bool declared;
     struct section_s *parent;
-    struct section_s *next;
     const struct file_list_s *file_list;
     struct optimizer_s *optimizer;
     struct linepos_s epoint;
@@ -73,8 +73,7 @@ extern struct section_s *find_this_section(const char *);
 extern void init_section(void);
 extern void destroy_section(void);
 extern void reset_section(struct section_s *);
-extern void sectionprint(void);
-extern void section_sizecheck(void);
+extern void section_sizecheck(const struct avltree_node *);
 extern struct section_s *current_section, root_section;
 extern struct section_address_s *current_address;
 #endif
