@@ -1,5 +1,5 @@
 /* ternary.c - Ternary Search Trees
-   $Id: ternary.c 1953 2019-08-31 14:49:15Z soci $
+   $Id: ternary.c 2896 2022-11-05 05:33:41Z soci $
 
    Copyright (C) 2001 Free Software Foundation, Inc.
 
@@ -26,17 +26,21 @@
 
 #ifdef DEBUG
 #define tern_free(tern) free(tern)
-#define tern_alloc() (ternary_node *)mallocx(sizeof(ternary_node))
+static MALLOC ternary_node *tern_alloc(void) {
+    ternary_node *r;
+    new_instance(&r);
+    return r;
+}
 #else
 static union tern_u {
     ternary_node tern;
     union tern_u *next;
-} *terns_free = NULL;
+} *terns_free;
 
 static struct terns_s {
     union tern_u terns[255];
     struct terns_s *next;
-} *terns = NULL;
+} *terns;
 
 static void tern_free(union tern_u *tern) {
     tern->next = terns_free;
@@ -46,7 +50,7 @@ static void tern_free(union tern_u *tern) {
 static union tern_u *terns_alloc(void) {
     size_t i;
     struct terns_s *old = terns;
-    terns = (struct terns_s *)mallocx(sizeof *terns);
+    new_instance(&terns);
     for (i = 0; i < 254; i++) {
         terns->terns[i].next = &terns->terns[i + 1];
     }
@@ -69,7 +73,7 @@ static MALLOC ternary_node *tern_alloc(void) {
 
 ternary_tree *ternary_insert(ternary_tree *pcurr, const uint8_t *s, const uint8_t *end)
 {
-    uchar_t spchar;
+    unichar_t spchar;
     ternary_tree curr;
   
     spchar = *s;
@@ -85,7 +89,7 @@ ternary_tree *ternary_insert(ternary_tree *pcurr, const uint8_t *s, const uint8_
             {
                 return &curr->eqkid;
             }
-            if (s == end) spchar = ~(uchar_t)0;
+            if (s == end) spchar = ~(unichar_t)0;
             else {
                 spchar = *s;
                 if ((spchar & 0x80) != 0) s += utf8in(s, &spchar); else s++;
@@ -110,7 +114,7 @@ ternary_tree *ternary_insert(ternary_tree *pcurr, const uint8_t *s, const uint8_
         if ((~spchar) == 0) {
             return pcurr;
         }
-        if (s == end) spchar = ~(uchar_t)0;
+        if (s == end) spchar = ~(unichar_t)0;
         else {
             spchar = *s;
             if ((spchar & 0x80) != 0) s += utf8in(s, &spchar); else s++;
@@ -132,33 +136,40 @@ void ternary_cleanup(ternary_tree p, ternary_free_fn_t f)
 /* Non-recursive find of a string in the ternary tree */
 void *ternary_search(const ternary_node *curr, const uint8_t *s, size_t *len)
 {
-    uchar_t spchar;
+    unichar_t spchar;
     const ternary_node *last = NULL;
-    size_t len2 = *len;
-    const uint8_t *end = s + len2;
+    const uint8_t *start = s, *end = s + *len, *prev = s;
 
     spchar = *s;
     if ((spchar & 0x80) != 0) s += utf8in(s, &spchar); else s++;
 
-    while (curr != NULL) {
+    do {
         if (spchar == curr->splitchar) {
             if ((~spchar) == 0) return (void *)curr->eqkid;
-            if (s == end) spchar = ~(uchar_t)0;
+            if (s == end) spchar = ~(unichar_t)0;
             else {
+                prev = s;
                 spchar = *s;
                 if ((spchar & 0x80) != 0) s += utf8in(s, &spchar); else s++;
             }
             curr = curr->eqkid;
             last = curr;
         } else curr = (spchar < curr->splitchar) ? curr->lokid : curr->hikid;
-    }
+    } while (curr != NULL);
     while (last != NULL && (~last->splitchar) != 0) {
         last = last->hikid;
     }
-    *len = len2 - (end - s) - 1;
+    *len = (size_t)(prev - start);
     return (last != NULL) ? (void *)last->eqkid : NULL;
 }
 
+void init_ternary(void) {
+#ifndef DEBUG
+    terns_free = NULL;
+    terns = NULL;
+#endif
+}
+
 void destroy_ternary(void)
 {
 #ifndef DEBUG
