@@ -1,5 +1,5 @@
 /*
-    $Id: typeobj.h 2122 2019-12-21 06:27:50Z soci $
+    $Id: typeobj.h 2896 2022-11-05 05:33:41Z soci $
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -31,10 +31,10 @@ typedef enum Truth_types {
 
 typedef enum Type_types {
     T_NONE, T_BOOL, T_BITS, T_INT, T_FLOAT, T_BYTES, T_STR, T_GAP, T_ADDRESS,
-    T_IDENT, T_ANONIDENT, T_ERROR, T_OPER, T_COLONLIST, T_TUPLE, T_LIST,
+    T_SYMBOL, T_ANONSYMBOL, T_ERROR, T_OPER, T_COLONLIST, T_TUPLE, T_LIST,
     T_DICT, T_MACRO, T_SEGMENT, T_UNION, T_STRUCT, T_MFUNC, T_CODE, T_LBL,
-    T_DEFAULT, T_REGISTER, T_FUNCTION, T_ADDRLIST, T_FUNCARGS, T_TYPE, T_LABEL,
-    T_NAMESPACE, T_MEMBLOCKS, T_FOLD
+    T_DEFAULT, T_REGISTER, T_FUNCTION, T_FUNCARGS, T_TYPE, T_LABEL,
+    T_NAMESPACE, T_MEMBLOCKS, T_FOLD, T_SFUNC, T_ENC
 } Type_types;
 
 typedef struct Type {
@@ -44,18 +44,20 @@ typedef struct Type {
     const char *name;
     struct Slot **slot;
     size_t length;
-    Obj *(*create)(Obj *, linepos_t) MUST_CHECK;
+    Obj *(*convert)(struct oper_s *) MUST_CHECK;
+    Obj *(*convert2)(struct oper_s *) MUST_CHECK;
     void (*destroy)(Obj *) FAST_CALL;
     void (*garbage)(Obj *, int) FAST_CALL;
     bool (*same)(const Obj *, const Obj *) FAST_CALL;
     Obj *(*truth)(Obj *, Truth_types, linepos_t) MUST_CHECK;
-    struct Error *(*hash)(Obj *, int *, linepos_t) MUST_CHECK;
+    Obj *(*hash)(Obj *, int *, linepos_t) MUST_CHECK;
     Obj *(*repr)(Obj *, linepos_t, size_t) MUST_CHECK;
     Obj *(*str)(Obj *, linepos_t, size_t) MUST_CHECK;
     Obj *(*calc1)(struct oper_s *) MUST_CHECK;
     Obj *(*calc2)(struct oper_s *) MUST_CHECK;
     Obj *(*rcalc2)(struct oper_s *) MUST_CHECK;
-    Obj *(*slice)(struct oper_s *, size_t) MUST_CHECK;
+    Obj *(*slice)(struct oper_s *, argcount_t) MUST_CHECK;
+    Obj *(*contains)(struct oper_s *) MUST_CHECK;
     struct Error *(*ival)(Obj *, ival_t *, unsigned int, linepos_t) MUST_CHECK;
     struct Error *(*uval)(Obj *, uval_t *, unsigned int, linepos_t) MUST_CHECK;
     struct Error *(*uval2)(Obj *, uval_t *, unsigned int, linepos_t) MUST_CHECK;
@@ -67,10 +69,16 @@ typedef struct Type {
     Obj *(*len)(struct oper_s *) MUST_CHECK;
     Obj *(*size)(struct oper_s *) MUST_CHECK;
     void (*getiter)(struct iter_s *);
+    void (*getriter)(struct iter_s *);
 } Type;
 
+enum { MAXIMUM_TYPE_LENGTH = 33 };
+
+#define Type(a) OBJ_CAST(Type, a)
+
+extern void init_type(void);
 extern void typeobj_init(void);
 extern void typeobj_names(void);
-extern void new_type(Type *, Type_types, const char *, size_t);
+extern Type *new_type(Type *, Type_types, const char *, size_t);
 
 #endif
