@@ -1,5 +1,5 @@
 /*
-    $Id: values.h 2108 2019-12-10 19:16:03Z soci $
+    $Id: values.h 2898 2022-11-05 08:08:41Z soci $
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -24,21 +24,18 @@
 struct Type;
 struct Obj;
 
-struct iter_s;
-typedef FAST_CALL MUST_CHECK struct Obj *(*iter_next_t)(struct iter_s *);
-
 struct iter_s {
     struct Obj *data;
     struct Obj *iter;
     size_t val;
-    iter_next_t next;
+    struct Obj *(*next)(struct iter_s *) FAST_CALL MUST_CHECK;
     size_t len;
 };
 
 extern FAST_CALL MALLOC struct Obj *val_alloc(const struct Type *);
 extern FAST_CALL void val_destroy(struct Obj *);
 extern FAST_CALL void val_replace(struct Obj **, struct Obj *);
-extern size_t val_print(struct Obj *, FILE *);
+extern size_t val_print(struct Obj *, FILE *, size_t);
 extern FAST_CALL void iter_destroy(struct iter_s *);
 
 extern void destroy_values(void);
