@@ -1,5 +1,5 @@
 /*
-    $Id: variables.c 2137 2020-01-02 00:52:17Z soci $
+    $Id: variables.c 2924 2022-12-22 08:52:34Z soci $
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -48,7 +48,7 @@
 #include "errorobj.h"
 #include "mfuncobj.h"
 
-static Label *lastlb = NULL;
+static Label *lastlb;
 
 #define EQUAL_COLUMN 16
 
@@ -70,12 +70,12 @@ struct context_stack_s {
 
 static struct context_stack_s context_stack;
 
+static void extend_context_stack(void) {
+    extend_array(&context_stack.stack, &context_stack.len, 8);
+}
+
 void push_context(Namespace *name) {
-    if (context_stack.p >= context_stack.len) {
-        context_stack.len += 8;
-        if (/*context_stack.len < 8 ||*/ context_stack.len > SIZE_MAX / sizeof *context_stack.stack) err_msg_out_of_memory(); /* overflow */
-        context_stack.stack = (struct cstack_s *)reallocx(context_stack.stack, context_stack.len * sizeof *context_stack.stack);
-    }
+    if (context_stack.p >= context_stack.len) extend_context_stack();
     context_stack.stack[context_stack.p].normal = ref_namespace(name);
     current_context = name;
     context_stack.stack[context_stack.p].cheap = cheap_context;
@@ -90,8 +90,8 @@ void push_dummy_context(void) {
 bool pop_context(void) {
     if (context_stack.p > 1 + context_stack.bottom) {
         struct cstack_s *c = &context_stack.stack[--context_stack.p];
-        val_destroy(&c->normal->v);
-        val_destroy(&cheap_context->v);
+        val_destroy(Obj(c->normal));
+        val_destroy(Obj(cheap_context));
         cheap_context = c->cheap;
         c = &context_stack.stack[context_stack.p - 1];
         current_context = c->normal;
@@ -101,11 +101,7 @@ bool pop_context(void) {
 }
 
 void push_context2(Namespace *name) {
-    if (context_stack.p >= context_stack.len) {
-        context_stack.len += 8;
-        if (/*context_stack.len < 8 ||*/ context_stack.len > SIZE_MAX / sizeof *context_stack.stack) err_msg_out_of_memory(); /* overflow */
-        context_stack.stack = (struct cstack_s *)reallocx(context_stack.stack, context_stack.len * sizeof *context_stack.stack);
-    }
+    if (context_stack.p >= context_stack.len) extend_context_stack();
     context_stack.stack[context_stack.p].normal = context_stack.stack[context_stack.p - 1].normal;
     context_stack.stack[context_stack.p - 1].normal = ref_namespace(name);
     context_stack.stack[context_stack.p].cheap = ref_namespace(name);
@@ -115,9 +111,9 @@ void push_context2(Namespace *name) {
 bool pop_context2(void) {
     if (context_stack.p > 1 + context_stack.bottom) {
         struct cstack_s *c = &context_stack.stack[--context_stack.p];
-        val_destroy(&context_stack.stack[context_stack.p - 1].normal->v);
+        val_destroy(Obj(context_stack.stack[context_stack.p - 1].normal));
         context_stack.stack[context_stack.p - 1].normal = c->normal;
-        val_destroy(&c->cheap->v);
+        val_destroy(Obj(c->cheap));
         return false;
     }
     return true;
@@ -127,10 +123,10 @@ void reset_context(void) {
     context_stack.bottom = 0;
     while (context_stack.p != 0) {
         struct cstack_s *c = &context_stack.stack[--context_stack.p];
-        val_destroy(&c->normal->v);
-        val_destroy(&c->cheap->v);
+        val_destroy(Obj(c->normal));
+        val_destroy(Obj(c->cheap));
     }
-    val_destroy(&cheap_context->v);
+    val_destroy(Obj(cheap_context));
     cheap_context = ref_namespace(root_namespace);
     push_context(root_namespace);
     root_namespace->backr = root_namespace->forwr = 0;
@@ -144,11 +140,7 @@ struct label_stack_s {
 static struct label_stack_s label_stack;
 
 static void push_label(Label *name) {
-    if (label_stack.p >= label_stack.len) {
-        label_stack.len += 8;
-        if (/*label_stack.len < 8 ||*/ label_stack.len > SIZE_MAX / sizeof(Label *)) err_msg_out_of_memory(); /* overflow */
-        label_stack.stack = (Label **)reallocx(label_stack.stack, label_stack.len * sizeof(Label *));
-    }
+    if (label_stack.p >= label_stack.len) extend_array(&label_stack.stack, &label_stack.len, 8);
     label_stack.stack[label_stack.p] = name;
     label_stack.p++;
 }
@@ -159,9 +151,8 @@ static void pop_label(void) {
 
 void get_namespaces(Mfunc *mfunc) {
     size_t i, len = context_stack.p - context_stack.bottom;
-    if (len > SIZE_MAX / sizeof *mfunc->namespaces) err_msg_out_of_memory(); /* overflow */
     mfunc->nslen = len;
-    mfunc->namespaces = (Namespace **)mallocx(len * sizeof *mfunc->namespaces);
+    new_array(&mfunc->namespaces, len);
     for (i = 0; i < len; i++) {
         mfunc->namespaces[i] = ref_namespace(context_stack.stack[context_stack.bottom + i].normal);
     }
@@ -183,8 +174,9 @@ static Label *namespace_update(Namespace *ns, Label *p) {
     size_t mask, hash, offs;
     if (ns->len * 3 / 2 >= ns->mask) {
         size_t i, max = (ns->data == NULL) ? 8 : (ns->mask + 1) << 1;
-        Label **n = (Label **)calloc(max, sizeof *n);
-        if (n == NULL) err_msg_out_of_memory();
+        Label **n;
+        new_array(&n, max);
+        memset(n, 0, max * sizeof *n);
         mask = max - 1;
         if (ns->data != NULL) {
             for (i = 0; i <= ns->mask; i++) if (ns->data[i] != NULL) {
@@ -340,7 +332,7 @@ Label *find_label2(const str_t *name, Namespace *context) {
     return namespace_lookup(context, &label);
 }
 
-struct anonident_s {
+struct anonsymbol_s {
     uint8_t dir, pad;
     uint8_t count[sizeof(uint32_t)];
 };
@@ -356,23 +348,30 @@ Label *find_label3(const str_t *name, Namespace *context, uint8_t strength) {
     return namespace_lookup3(context, &label);
 }
 
-Label *find_anonlabel(int32_t count) {
+Label *find_anonlabel(ssize_t count) {
     size_t p = context_stack.p;
     Namespace *context;
     Label label, *c;
-    struct anonident_s anonident;
+    struct anonsymbol_s anonsymbol;
 
-    anonident.dir = (count >= 0) ? '+' : '-';
-    anonident.pad = 0;
+    anonsymbol.dir = (count >= 0) ? '+' : '-';
+    anonsymbol.pad = 0;
 
-    label.cfname.data = (const uint8_t *)&anonident;
+    label.cfname.data = (const uint8_t *)&anonsymbol;
 
     while (context_stack.bottom < p) {
+        uint32_t count2;
         context = context_stack.stack[--p].normal;
-        uint32_t count2 = (uint32_t)((int32_t)((count >= 0) ? context->forwr : context->backr) + count);
+        if (count < 0) {
+            if (context->backr < -(size_t)count) continue;
+            count2 = context->backr - -(uint32_t)count;
+        } else {
+            count2 = context->forwr + (uint32_t)count;
+            if (count2 < (size_t)count) continue;
+        }
         label.cfname.len = 2;
         while (count2 != 0) {
-            anonident.count[label.cfname.len - 2] = count2;
+            anonsymbol.count[label.cfname.len - 2] = (uint8_t)count2;
             label.cfname.len++;
             count2 >>= 8;
         } 
@@ -384,18 +383,27 @@ Label *find_anonlabel(int32_t count) {
     return NULL;
 }
 
-Label *find_anonlabel2(int32_t count, Namespace *context) {
+Label *find_anonlabel2(ssize_t count, Namespace *context) {
     Label label;
-    uint32_t count2 = (uint32_t)((int32_t)((count >= 0) ? 0 : context->backr) + count);
-    struct anonident_s anonident;
+    uint32_t count2;
+    struct anonsymbol_s anonsymbol;
 
-    anonident.dir = (count >= 0) ? '+' : '-';
-    anonident.pad = 0;
+    if (count < 0) {
+        if (context->backr < -(size_t)count) return NULL;
+        count2 = context->backr - -(uint32_t)count;
+        anonsymbol.dir = '-';
+    } else {
+        count2 = context->forwr + (uint32_t)count;
+        if (count2 < (size_t)count) return NULL;
+        anonsymbol.dir = '+';
+    }
 
-    label.cfname.data = (const uint8_t *)&anonident;
+    anonsymbol.pad = 0;
+
+    label.cfname.data = (const uint8_t *)&anonsymbol;
     label.cfname.len = 2;
     while (count2 != 0) {
-        anonident.count[label.cfname.len - 2] = count2;
+        anonsymbol.count[label.cfname.len - 2] = (uint8_t)count2;
         label.cfname.len++;
         count2 >>= 8;
     } 
@@ -405,9 +413,9 @@ Label *find_anonlabel2(int32_t count, Namespace *context) {
 }
 
 /* --------------------------------------------------------------------------- */
-Label *new_label(const str_t *name, Namespace *context, uint8_t strength, bool *exists, const struct file_list_s *cflist) {
+Label *new_label(const str_t *name, Namespace *context, uint8_t strength, const struct file_list_s *cflist) {
     Label *b;
-    if (lastlb == NULL) lastlb = (Label *)val_alloc(LABEL_OBJ);
+    if (lastlb == NULL) lastlb = Label(val_alloc(LABEL_OBJ));
 
     if (name->len > 1 && name->data[1] == 0) lastlb->cfname = *name;
     else str_cfcpy(&lastlb->cfname, name);
@@ -417,8 +425,8 @@ Label *new_label(const str_t *name, Namespace *context, uint8_t strength, bool *
     b = namespace_update(context, lastlb);
 
     if (b == NULL) { /* new label */
-        if ((size_t)(name->data - cflist->file->data) < cflist->file->len) lastlb->name = *name;
-        else str_cpy(&lastlb->name, name);
+        if (not_in_file(name->data, cflist->file)) str_cpy(&lastlb->name, name);
+        else lastlb->name = *name;
         if (lastlb->cfname.data != name->data) str_cfcpy(&lastlb->cfname, NULL);
         else lastlb->cfname = lastlb->name;
         lastlb->file_list = cflist;
@@ -426,21 +434,19 @@ Label *new_label(const str_t *name, Namespace *context, uint8_t strength, bool *
         lastlb->update_after = false;
         lastlb->usepass = 0;
         lastlb->fwpass = 0;
+        lastlb->value = NULL;
         lastlb->defpass = pass;
-        *exists = false;
         b = lastlb;
         lastlb = NULL;
-        return b;
     }
-    *exists = true;
     return b;
 }
 
 void label_move(Label *label, const str_t *name, const struct file_list_s *cflist) {
     bool cfsame = (label->cfname.data == label->name.data);
-    if ((size_t)(label->name.data - label->file_list->file->data) < label->file_list->file->len) {
-        if ((size_t)(name->data - cflist->file->data) < cflist->file->len) label->name = *name;
-        else str_cpy(&label->name, name);
+    if (!not_in_file(label->name.data, label->file_list->file)) {
+        if (not_in_file(name->data, cflist->file)) str_cpy(&label->name, name);
+        else label->name = *name;
     }
     if (cfsame) {
         label->cfname = label->name;
@@ -482,18 +488,18 @@ void unused_check(Namespace *names) {
         if (!key2->owner) continue;
         switch (o->obj->type) {
         case T_CODE:
-            ns = ((Code *)o)->names;
+            ns = Code(o)->names;
             break;
         case T_NAMESPACE:
-            ns = (Namespace *)o;
+            ns = Namespace(o);
             break;
         case T_MFUNC:
             {
-                Mfunc *mfunc = (Mfunc *)o;
+                Mfunc *mfunc = Mfunc(o);
                 List *lst = mfunc->inamespaces;
                 size_t i;
                 for (i = 0; i < lst->len; i++) {
-                    ns = (Namespace *)lst->data[i];
+                    ns = Namespace(lst->data[i]);
                     if (ns->len != 0) unused_check(ns);
                 }
                 ns = mfunc->names;
@@ -529,28 +535,7 @@ static void labelname_print(const Label *l, FILE *flab, char d) {
     printable_print2(l->name.data, flab, l->name.len);
 }
 
-struct vice_address_s {
-    address_t addr;
-    struct avltree_node node;
-};
-
-static struct vice_addresses_s {
-    struct vice_address_s list[256];
-    struct vice_addresses_s *next;
-} *vice_addresses = NULL;
-
-static struct avltree vice_filter;
-static struct vice_address_s *lastvc;
-static int vice_addrp;
-
-static FAST_CALL int duplicate_compare(const struct avltree_node *aa, const struct avltree_node *bb)
-{
-    const struct vice_address_s *a = cavltree_container_of(aa, struct vice_address_s, node);
-    const struct vice_address_s *b = cavltree_container_of(bb, struct vice_address_s, node);
-    return (int)a->addr - (int)b->addr;
-}
-
-static void labelprint2(Namespace *names, FILE *flab, int labelmode) {
+static void labelprint2(Namespace *names, FILE *flab, Symbollist_types labelmode) {
     size_t n, ln;
 
     if (names->len == 0) return;
@@ -566,7 +551,7 @@ static void labelprint2(Namespace *names, FILE *flab, int labelmode) {
         case T_UNION:
         case T_STRUCT: continue;
         case T_CODE:
-            if (((Code *)l->value)->pass != ((Code *)l->value)->apass) continue;
+            if (Code(l->value)->pass != Code(l->value)->apass) continue;
             break;
         default:break;
         }
@@ -595,19 +580,10 @@ static void labelprint2(Namespace *names, FILE *flab, int labelmode) {
                 uval_t uv;
                 Error *err = val->obj->uval(val, &uv, 24, &epoint);
                 if (err == NULL) {
-                    lastvc->addr = uv;
-                    if (avltree_insert(&lastvc->node, &vice_filter, duplicate_compare) != NULL) continue;
-                    if (vice_addrp == 255) {
-                        struct vice_addresses_s *old = vice_addresses;
-                        vice_addresses = (struct vice_addresses_s *)mallocx(sizeof *vice_addresses);
-                        vice_addresses->next = old;
-                        vice_addrp = 0;
-                    } else vice_addrp++;
-                    lastvc = &vice_addresses->list[vice_addrp];
                     fprintf(flab, "al %" PRIx32 " .", uv & 0xffffff);
                     labelname_print(l, flab, ':');
                     putc('\n', flab);
-                } else val_destroy(&err->v);
+                } else val_destroy(Obj(err));
             }
             if (l->owner) {
                 Namespace *ns = get_namespace(val);
@@ -619,9 +595,10 @@ static void labelprint2(Namespace *names, FILE *flab, int labelmode) {
             }
         } else {
             Obj *val = l->value;
-            Str *str = (Str *)val->obj->repr(val, NULL, SIZE_MAX);
-            if (str == NULL) continue;
-            if (str->v.obj == STR_OBJ) {
+            val = val->obj->repr(val, NULL, SIZE_MAX);
+            if (val == NULL) continue;
+            if (val->obj == STR_OBJ) {
+                const Str *str = Str(val);
                 size_t len = printable_print2(l->name.data, flab, l->name.len);
                 padding(len, EQUAL_COLUMN, flab);
                 if (l->constant) fputs("= ", flab);
@@ -629,14 +606,14 @@ static void labelprint2(Namespace *names, FILE *flab, int labelmode) {
                 printable_print2(str->data, flab, str->len);
                 putc('\n', flab);
             }
-            val_destroy(&str->v);
+            val_destroy(val);
         }
     }
     names->len = ln;
 }
 
-static inline const uint8_t *get_line(const struct file_s *file, size_t line) {
-    return &file->data[file->line[line - 1]];
+static inline const uint8_t *get_line(const struct file_s *file, linenum_t line) {
+    return &file->source.data[file->line[line - 1]];
 }
 
 static void labeldump(Namespace *names, FILE *flab) {
@@ -650,19 +627,21 @@ static void labeldump(Namespace *names, FILE *flab) {
 
         if (l2 == NULL) continue;
         if (l2->name.len < 2 || l2->name.data[1] != 0) {
-            Str *val = (Str *)l2->value->obj->repr(l2->value, NULL, SIZE_MAX);
+            Obj *val = l2->value;
+            val = val->obj->repr(val, NULL, SIZE_MAX);
             if (val != NULL) {
-                if (val->v.obj == STR_OBJ) {
+                if (val->obj == STR_OBJ) {
+                    const Str *str = Str(val);
                     const struct file_s *file = l2->file_list->file;
                     linepos_t epoint = &l2->epoint;
-                    printable_print((const uint8_t *)file->realname, flab);
+                    printable_print((const uint8_t *)file->name, flab);
                     fprintf(flab, ":%" PRIuline ":%" PRIlinepos ": ", epoint->line, ((file->encoding == E_UTF8) ? (linecpos_t)calcpos(get_line(file, epoint->line), epoint->pos) : epoint->pos) + 1);
                     labelname_print(l2, flab, '.');
                     fputs(l2->constant ? " = " : " := ", flab);
-                    printable_print2(val->data, flab, val->len);
+                    printable_print2(str->data, flab, str->len);
                     putc('\n', flab);
                 }
-                val_destroy(&val->v);
+                val_destroy(val);
             }
         }
         if (!l2->owner) continue;
@@ -693,7 +672,7 @@ static Namespace *find_space(const char *here, bool use) {
     do {
         labelname.data = pline + lpoint.pos; labelname.len = get_label(labelname.data);
         if (labelname.len == 0) return NULL;
-        lpoint.pos += labelname.len;
+        lpoint.pos += (linecpos_t)labelname.len;
         l = find_label2(&labelname, space);
         if (l == NULL) return NULL;
         space = get_namespace(l->value);
@@ -710,17 +689,18 @@ static Namespace *find_space(const char *here, bool use) {
     return space;
 }
 
-bool labelprint(const struct symbol_output_s *output, bool append) {
+void labelprint(const struct symbol_output_s *output) {
     FILE *flab;
     struct linepos_s nopoint = {0, 0};
     int err;
     Namespace *space;
 
-    flab = dash_name(output->name) ? stdout : file_open(output->name, append ? "at" : "wt");
+    flab = dash_name(output->name) ? stdout : fopen_utf8(output->name, output->append ? "at" : "wt");
     if (flab == NULL) {
-        err_msg_file(ERROR_CANT_WRTE_LBL, output->name, &nopoint);
-        return true;
+        err_msg_file2(ERROR_CANT_WRTE_LBL, output->name);
+        return;
     }
+    if (flab == stdout && fflush(flab) != 0) setvbuf(flab, NULL, _IOLBF, 1024);
     clearerr(flab); errno = 0;
     label_stack.stack = NULL;
     label_stack.p = label_stack.len = 0;
@@ -732,18 +712,6 @@ bool labelprint(const struct symbol_output_s *output, bool append) {
         err_msg2(ERROR____LABEL_ROOT, &labelname, &nopoint);
     } else if (output->mode == LABEL_DUMP) {
         labeldump(space, flab);
-    } else if (output->mode == LABEL_VICE || output->mode == LABEL_VICE_NUMERIC) {
-        vice_addresses = (struct vice_addresses_s *)mallocx(sizeof *vice_addresses);
-        vice_addresses->next = NULL;
-        vice_addrp = 0;
-        avltree_init(&vice_filter);
-        lastvc = &vice_addresses->list[vice_addrp];
-        labelprint2(space, flab, output->mode);
-        while (vice_addresses != NULL) {
-            struct vice_addresses_s *old = vice_addresses;
-            vice_addresses = vice_addresses->next;
-            free(old);
-        }
     } else {
         labelprint2(space, flab, output->mode);
     }
@@ -751,10 +719,8 @@ bool labelprint(const struct symbol_output_s *output, bool append) {
     err = ferror(flab);
     err |= (flab != stdout) ? fclose(flab) : fflush(flab);
     if (err != 0 && errno != 0) {
-        err_msg_file(ERROR_CANT_WRTE_LBL, output->name, &nopoint);
-        return true;
+        err_msg_file2(ERROR_CANT_WRTE_LBL, output->name);
     }
-    return false;
 }
 
 void ref_labels(void) {
@@ -781,23 +747,20 @@ void ref_labels(void) {
             default:break;
             }
             if (l != namespace_lookup(space, l)) continue;
-            if (l->value->obj == ERROR_OBJ) err_msg_output((Error *)l->value);
+            if (l->value->obj == ERROR_OBJ) err_msg_output(Error(l->value));
             l->ref = true;
             l->usepass = pass;
         }
     }
 }
 
-static struct file_list_s dummy_cflist;
-
-void new_builtin(const char *ident, Obj *val) {
+void new_builtin(const char *symbol, Obj *val) {
     struct linepos_s nopoint = {0, 0};
     str_t name;
     Label *label;
-    bool label_exists;
-    name.len = strlen(ident);
-    name.data = (const uint8_t *)ident;
-    label = new_label(&name, builtin_namespace, 0, &label_exists, &dummy_cflist);
+    name.len = strlen(symbol);
+    name.data = (const uint8_t *)symbol;
+    label = new_label(&name, builtin_namespace, 0, dummy_file_list);
     label->constant = true;
     label->owner = true;
     label->value = val;
@@ -806,14 +769,15 @@ void new_builtin(const char *ident, Obj *val) {
 
 void init_variables(void)
 {
+    enum { BUILTIN_SIZE = 128 };
     struct linepos_s nopoint = {0, 0};
-    static struct file_s cfile;
-
-    cfile.data = (uint8_t *)0;
-    cfile.len = SIZE_MAX;
-    dummy_cflist.file = &cfile;
 
     builtin_namespace = new_namespace(NULL, &nopoint);
+    builtin_namespace->data = allocate_array(Label *, BUILTIN_SIZE);
+    if (builtin_namespace->data != NULL) {
+        memset(builtin_namespace->data, 0, BUILTIN_SIZE * sizeof *builtin_namespace->data);
+        builtin_namespace->mask = BUILTIN_SIZE - 1;
+    }
     root_namespace = new_namespace(NULL, &nopoint);
     cheap_context = ref_namespace(root_namespace);
 
@@ -840,21 +804,21 @@ void init_variables(void)
 void destroy_lastlb(void) {
     if (lastlb != NULL) {
         lastlb->v.obj = NONE_OBJ;
-        val_destroy(&lastlb->v);
+        val_destroy(Obj(lastlb));
         lastlb = NULL;
     }
 }
 
 void destroy_variables(void)
 {
-    val_destroy(&builtin_namespace->v);
-    val_destroy(&root_namespace->v);
-    val_destroy(&cheap_context->v);
+    val_destroy(Obj(builtin_namespace));
+    val_destroy(Obj(root_namespace));
+    val_destroy(Obj(cheap_context));
     destroy_lastlb();
     while (context_stack.p != 0) {
         struct cstack_s *c = &context_stack.stack[--context_stack.p];
-        val_destroy(&c->normal->v);
-        val_destroy(&c->cheap->v);
+        val_destroy(Obj(c->normal));
+        val_destroy(Obj(c->cheap));
     }
     free(context_stack.stack);
 }
