@@ -1,5 +1,5 @@
 /*
-    $Id: wait_e.h 2131 2019-12-30 19:02:20Z soci $
+    $Id: wait_e.h 2886 2022-11-01 08:01:49Z soci $
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -19,10 +19,12 @@
 #ifndef WAIT_E_H
 #define WAIT_E_H
 typedef enum Wait_types {
-    W_NONE, W_ENDM, W_ENDM3, W_BEND, W_BEND2, W_HERE, W_HERE2, W_ENDU, W_ENDU2,
-    W_ENDU3, W_ENDS, W_ENDS2, W_ENDS3, W_ENDC, W_ENDP, W_ENDP2, W_NEXT,
-    W_NEXT2, W_NEXT3, W_SEND, W_SEND2, W_PEND, W_FI, W_FI2, W_ENDF, W_ENDF3,
-    W_SWITCH, W_SWITCH2, W_WEAK, W_WEAK2, W_ENDN, W_ENDN2, W_ENDV, W_ENDV2,
-    W_ENDWITH, W_ENDWITH2
+    W_NONE, W_ENDMACRO, W_ENDMACRO2, W_ENDSEGMENT, W_ENDSEGMENT2, W_BEND,
+    W_BEND2, W_HERE, W_HERE2, W_ENDU, W_ENDU2, W_ENDU3, W_ENDS, W_ENDS2,
+    W_ENDS3, W_ENDC, W_ENDP, W_ENDP2, W_ENDFOR, W_ENDFOR2, W_ENDFOR3,
+    W_ENDREPT, W_ENDREPT2, W_ENDREPT3, W_ENDWHILE, W_ENDWHILE2, W_ENDWHILE3,
+    W_SEND, W_SEND2, W_PEND, W_FI, W_FI2, W_ENDF, W_ENDF2, W_ENDF3, W_SWITCH,
+    W_SWITCH2, W_WEAK, W_WEAK2, W_ENDN, W_ENDN2, W_ENDV, W_ENDV2, W_ENDWITH,
+    W_ENDWITH2, W_ENDENCODE, W_ENDENCODE2
 } Wait_types;
 #endif
